#!/usr/bin/env python3
############################################################
# Program is part of MintPy                                #
# Copyright (c) 2013, Zhang Yunjun, Heresh Fattahi         #
# Author: Heresh Fattahi, 2013                             #
############################################################


import getopt
import os
import sys

import h5py
import matplotlib.pyplot as plt
import numpy as np


def usage():
    print("""
*****************************************************************************************
   Generating multiple profiles(each profile includes seeveral transects [specified by -n])
   perpendicular to a Fault . Fault is a path specified by lat and lon coordinates.

   Usage:
       -n number of transects used to generate one profile
       -d distance [in pixel] between individual transects to generate one profile
       -F a txt file including the fault coordinates (first column lon , second column: lat)
       -p flip profile left-right (yes or no) [default: no]
       -u flip up - down [default: no]
       -g gps_file (if exists)
       -S source of GPS velocities (usgs,cmm4,mintpy)
       -G gps stations to compare with InSAR  (all,insar,profile)
          "all": all gps stations is projected to the profile
          "insar": same as all but limited to the area covered by insar
          "profile": only those gps stations which are in the profile area]

       -x lower bound to display in x direction
       -X higher bound to display in x direction
       -l lower bound to display in y direction
       -h higher bound to display in y direction
       -I display InSAR velocity [on] or off
       -A display Average InSAR velocity [on] or off
       -U display Standard deviation of the InSAR velocity [on] or off
       -E Export the generated transect to a matlab file [off] or on
       -W Length of a profile
       -D Distance between two consequent average profile

   Example:
       multi_transect.py -f geo_velocity_masked.h5 -n 50 -d 1 -W 10 -D 2 -F Chaman_fault.txt

********************************************************************************************
    """)
    return


def dms2d(Coord):
    d, m, s = Coord.split(' ')
    d = float(d)
    m = float(m)
    s = float(s)
    d = d+m/60.+s/3600.
    return d


def gps_to_LOS(Ve, Vn, theta, heading):

    unitVec = [np.cos(heading)*np.sin(theta), -np.sin(theta)
               * np.sin(heading), np.cos(theta)]

    gpsLOS = unitVec[0]*Ve+unitVec[1]*Vn

    return gpsLOS


def check_st_in_box(x, y, x0, y0, x1, y1, X0, Y0, X1, Y1):

    m1 = float(y1-y0)/float(x1-x0)
    c1 = float(y0-m1*x0)

    m2 = float(Y1-Y0)/float(X1-X0)
    c2 = float(Y0-m2*X0)

    m3 = float(y0-Y0)/float(x0-X0)
    c3 = float(Y0-m3*X0)

    m4 = float(y1-Y1)/float(x1-X1)
    c4 = float(Y1-m4*X1)

    yy1 = m1*x+c1
    yy2 = m2*x+c2

    yy = [yy1, yy2]
    yy.sort()

    xx3 = (y-c3)/m3
    xx4 = (y-c4)/m4
    xx = [xx3, xx4]
    xx.sort()
    if y >= yy[0] and y <= yy[1] and x >= xx[0] and x <= xx[1]:
        Check_result = 'True'
    else:
        Check_result = 'False'

    return Check_result


def check_st_in_box2(x, y, x0, y0, x1, y1, X0, Y0, X1, Y1):

    m1, c1 = line(x0, y0, x1, y1)
    m2, c2 = line(X0, Y0, X1, Y1)
    m3, c3 = line(x0, y0, X0, Y0)
    m4, c4 = line(x1, y1, X1, Y1)

    D1 = np.sqrt((x1-x0)**2+(y1-y0)**2)
    D2 = np.sqrt((X0-x0)**2+(Y0-y0)**2)

    d1 = dist_point_from_line(m1, c1, x, y, 1, 1)
    d2 = dist_point_from_line(m2, c2, x, y, 1, 1)

    d3 = dist_point_from_line(m3, c3, x, y, 1, 1)
    d4 = dist_point_from_line(m4, c4, x, y, 1, 1)

    if np.round(d1+d2) == np.round(D2) and np.round(d3+d4) == np.round(D1):
        Check_result = 'True'
    else:
        Check_result = 'False'

    return Check_result


def line(x0, y0, x1, y1):
    m = float(y1-y0)/float(x1-x0)
    c = float(y0-m*x0)
    return m, c


def dist_point_from_line(m, c, x, y, dx, dy):
    # finds the distance of a point at x ,y xoordinate
    # from a line with Y =  mX +c

    d = np.sqrt((((x+m*y-m*c)/(m**2+1)-x)*dx)**2 +
                ((m*(x+m*y-m*c)/(m**2+1)+c-y)*dy)**2)
    #a=m;b=-1;
    #d=np.abs(a*x+b*y+c)/np.sqrt(a**2+b**2)
    return d


def get_intersect(m, c, x, y):

    xp = (x+m*y-m*c)/(m**2+1)
    yp = m*(x+m*y-m*c)/(m**2+1)+c
    return xp, yp


def readGPSfile(gpsFile, gps_source):

    if gps_source in ['cmm4', 'CMM4']:

        gpsData = np.loadtxt(gpsFile, usecols=(1, 2, 3, 4, 5, 6, 7, 8, 9, 10))
        Stations = np.loadtxt(gpsFile, dtype=str, usecols=(0, 1))[:, 0]

        St = []
        Lon = []
        Lat = []
        Ve = []
        Vn = []
        Se = []
        Sn = []
        for i in range(gpsData.shape[0]):
            if 'GPS' in Stations[i]:
                Lat.append(gpsData[i, 0])
                Lon.append(gpsData[i, 1])
                Ve.append(gpsData[i, 2])
                Se.append(gpsData[i, 3])
                Vn.append(gpsData[i, 4])
                Sn.append(gpsData[i, 5])
                St.append(Stations[i])

    elif gps_source == 'mintpy':

        gpsData = np.loadtxt(gpsFile, usecols=(1, 2, 3, 4, 5, 6))  # ,7,8,9))
        Stations = np.loadtxt(gpsFile, dtype=str, usecols=(0, 1))[:, 0]

        St = []
        Lon = []
        Lat = []
        Ve = []
        Vn = []
        Se = []
        Sn = []

        for i in range(gpsData.shape[0]):
            Lon.append(gpsData[i, 0])
            Lat.append(gpsData[i, 1])
            Ve.append(gpsData[i, 2])
            Vn.append(gpsData[i, 3])
            Se.append(gpsData[i, 4])
            Sn.append(gpsData[i, 5])
            St.append(Stations[i])

    elif gps_source in ['usgs', 'USGS']:

        gpsData_Hz = np.loadtxt(gpsFile, usecols=(0, 1, 2, 3, 4, 5, 6))
        gpsData_up = np.loadtxt(gpsFile, usecols=(8, 9))
        gpsData = np.hstack((gpsData_Hz, gpsData_up))
        Stations = np.loadtxt(gpsFile, dtype=str, usecols=(7, 8))[:, 0]

        St = []
        Lon = []
        Lat = []
        Ve = []
        Vn = []
        Se = []
        Sn = []

        for i in range(gpsData.shape[0]):
            Lat.append(gpsData[i, 0])
            Lon.append(gpsData[i, 1])
            Vn.append(gpsData[i, 2])
            Ve.append(gpsData[i, 3])
            Sn.append(gpsData[i, 4])
            Se.append(gpsData[i, 5])
            St.append(Stations[i])

    return list(St), Lat, Lon, Ve, Se, Vn, Sn


def redGPSfile(gpsFile):
    gpsData_Hz = np.loadtxt(gpsFile, usecols=(0, 1, 2, 3, 4, 5, 6))
    gpsData_up = np.loadtxt(gpsFile, usecols=(8, 9))
    gpsData = np.hstack((gpsData_Hz, gpsData_up))
    Stations = np.loadtxt(gpsFile, dtype=str, usecols=(7, 8))[:, 0]
    return list(Stations), gpsData


def redGPSfile_cmm4(gpsFile):
    #gpsData = np.loadtxt(gpsFile,usecols = (1,2,3,4,5,6,7,8,9,10))
    gpsData = np.loadtxt(gpsFile, usecols=(1, 2, 3, 4, 5, 6, 7, 8, 9))
    #gpsData_up = np.loadtxt(gpsFile,usecols = (8,9))
    #gpsData=np.hstack((gpsData_Hz,gpsData_up))
    Stations = np.loadtxt(gpsFile, dtype=str, usecols=(0, 1))[:, 0]
    return list(Stations), gpsData


def nearest(x, tbase, xstep):
    """ find nearest neighbour """
    dist = np.sqrt((tbase - x)**2)
    if min(dist) <= np.abs(xstep):
        indx = dist == min(dist)
    else:
        indx = []
    return indx


def find_row_column(Lon, Lat, lon, lat, lon_step, lat_step):
    ################################################
    # finding row and column numbers of the GPS point

    idx = nearest(Lon, lon, lon_step)
    idy = nearest(Lat, lat, lat_step)
    if idx != [] and idy != []:
        IDX = np.where(idx == True)[0][0]
        IDY = np.where(idy == True)[0][0]
    else:
        IDX = np.nan
        IDY = np.nan
    return IDY, IDX

################################################


def get_lat_lon(h5file):
    k = list(h5file.keys())
    Length = float(h5file[k[0]].attrs['LENGTH'])
    Width = float(h5file[k[0]].attrs['WIDTH'])
    ullon = float(h5file[k[0]].attrs['X_FIRST'])
    ullat = float(h5file[k[0]].attrs['Y_FIRST'])
    lon_step = float(h5file[k[0]].attrs['X_STEP'])
    lat_step = float(h5file[k[0]].attrs['Y_STEP'])

    #Length,Width = np.shape(insarData)

    lllat = ullat+(Length-1)*lat_step
    urlon = ullon+(Width-1)*lon_step
    lat = np.linspace(ullat, lllat, Length)
    lon = np.linspace(ullon, urlon, Width)

    lon_all = np.tile(lon, (Length, 1))
    lat_all = np.tile(lat, (Width, 1)).T

    #lat=np.arange(ullat,lllat,lat_step)
    #lon=np.arange(ullon,urlon,lon_step)
    return lat, lon, lat_step, lon_step, lat_all, lon_all


def nanmean(data, **args):
    return np.ma.filled(np.ma.masked_array(data, np.isnan(data)).mean(**args), fill_value=np.nan)


def nanstd(data, **args):
    return np.ma.filled(np.ma.masked_array(data, np.isnan(data)).std(**args), fill_value=np.nan)


def get_transect(z, x0, y0, x1, y1):

    length = int(np.hypot(x1-x0, y1-y0))
    x, y = np.linspace(x0, x1, length), np.linspace(y0, y1, length)
    zi = z[y.astype(np.int), x.astype(np.int)]
    return zi


def get_start_end_point(Xf0, Yf0, Xf1, Yf1, L, dx, dy):

    L = L*1000/dx
    m, c = line(Xf0, Yf0, Xf1, Yf1)
    mp = -1./m

    a = np.arctan(1./mp)
    xs = L*np.sin(a)+Xf0
    ys = L*np.cos(a)+Yf0

    xe = -L*np.sin(a)+Xf0
    ye = -L*np.cos(a)+Yf0

    return int(ys), int(xs), int(ye), int(xe)


def point_with_distance_from_line(Xf0, Yf0, Xf1, Yf1, L):
    m, c = line(Xf0, Yf0, Xf1, Yf1)
    mp = -1./m

    a = np.arctan(1./mp)
    x = L*np.sin(a)+Xf0
    y = L*np.cos(a)+Yf0
    return x, y


def point_on_line_with_distance_from_beginning(Xf0, Yf0, Xf1, Yf1, L):
    m, c = line(Xf0, Yf0, Xf1, Yf1)

    a = np.arctan(1./m)
    x = L*np.sin(a)+Xf0
    y = L*np.cos(a)+Yf0
    return x, y


def read_fault_coords(Fault_coord_file, Dp):

    Coords = np.loadtxt(Fault_coord_file)
    Dp = Dp*180.0/np.pi/6375.0
    Np = Coords.shape[0]

    Fault_lon = []
    Fault_lat = []
    Fault_lon.append(Coords[0, 0])
    Fault_lat.append(Coords[0, 1])

    for i in range(Np-1):
        # Fault_lon.append(Coords[i,0])
        # Fault_lat.append(Coords[i,1])
        Xf0 = Coords[i, 0]
        Yf0 = Coords[i, 1]
        Xf1 = Coords[i+1, 0]
        Yf1 = Coords[i+1, 1]
        DY = Yf1-Yf0
        DX = Xf1-Xf0
        Fault_Segment = np.hypot(DX, DY)

        if Fault_Segment <= Dp:

            Fault_lon.append(Coords[i+1, 0])
            Fault_lat.append(Coords[i+1, 1])

        else:
            L = Dp
            ii = 0
            while Fault_Segment > L:
                ii = ii+1
                x, y = point_on_line_with_distance_from_beginning(
                    Xf0, Yf0, Xf1, Yf1, Dp*ii)
                Dx = y-Yf0
                Dy = x-Xf0
                L = np.hypot(Dx, Dy)
                if L <= Fault_Segment:
                    Fault_lon.append(x)
                    Fault_lat.append(y)
                print(ii)
    return Fault_lon, Fault_lat


#####################################################################
def main(argv=None):
    ntrans = 1
    save_to_mat = 'off'
    flip_profile = 'no'
    which_gps = 'all'
    flip_updown = 'yes'
    incidence_file = 'incidence_file'
    display_InSAR = 'on'
    display_Average = 'on'
    display_Standard_deviation = 'on'

    try:
        opts, args = getopt.getopt(
            argv, "f:s:e:n:d:g:l:h:r:L:F:p:u:G:S:i:I:A:U:E:D:W:x:X:")

    except getopt.GetoptError:
        usage()
        sys.exit(1)

    for opt, arg in opts:
        if opt == '-f':
            velocityFile = arg
        elif opt == '-s':
            pnt1 = arg.split(',')
            y0 = int(pnt1[0])
            x0 = int(pnt1[1])
        elif opt == '-e':
            pnt2 = arg.split(',')
            y1 = int(pnt2[0])
            x1 = int(pnt2[1])
        elif opt == '-n':
            ntrans = int(arg)
        elif opt == '-d':
            dp = float(arg)
        elif opt == '-g':
            gpsFile = arg
        elif opt == '-r':
            refStation = arg
        elif opt == '-i':
            incidence_file = arg
        elif opt == '-L':
            stationsList = arg.split(',')
        elif opt == '-F':
            Fault_coord_file = arg
        elif opt == '-p':
            flip_profile = arg
        elif opt == '-u':
            flip_updown = arg
            print(flip_updown)
        elif opt == '-G':
            which_gps = arg
        elif opt == '-S':
            gps_source = arg
        elif opt == '-l':
            lbound = float(arg)
        elif opt == '-I':
            display_InSAR = arg
        elif opt == '-A':
            display_Average = arg
        elif opt == '-U':
            display_Standard_deviation = arg
        elif opt == '-E':
            save_to_mat = arg
        elif opt == '-h':
            hbound = float(arg)
        elif opt == '-D':
            Dp = float(arg)
        elif opt == '-W':
            profile_Length = float(arg)
        elif opt == '-x':
            x_lbound = float(arg)
        elif opt == '-X':
            x_hbound = float(arg)

    try:
        h5file = h5py.File(velocityFile, 'r')
    except:
        usage()
        sys.exit(1)

    k = list(h5file.keys())
    dset = h5file[k[0]].get(k[0])
    z = dset[0:dset.shape[0], 0:dset.shape[1]]
    dx = float(h5file[k[0]].attrs['X_STEP'])*6375000.0*np.pi/180.0
    dy = float(h5file[k[0]].attrs['Y_STEP'])*6375000.0*np.pi/180.0

    #############################################################################

    try:
        lat, lon, lat_step, lon_step, lat_all, lon_all = get_lat_lon(h5file)
    except:
        print('radar coordinate')

    Fault_lon, Fault_lat = read_fault_coords(Fault_coord_file, Dp)

    # Fault_lon=[66.40968453947265,66.36000186563085,66.31103920134248]
    # Fault_lat=[30.59405079532564,30.51565960186412,30.43928430936202]

    Num_profiles = len(Fault_lon)-1
    print('*********************************************')
    print('*********************************************')
    print('Number of profiles to be generated: '+str(Num_profiles))
    print('*********************************************')
    print('*********************************************')

    for Np in range(Num_profiles):
        FaultCoords = [Fault_lat[Np], Fault_lon[Np],
                       Fault_lat[Np+1], Fault_lon[Np+1]]
        print('%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%')
        print('%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%')
        print('')
        print('Profile '+str(Np) + ' [of total '+str(Num_profiles)+']')
        print('')

        try:
            #  Lat0 = dms2d(FaultCoords[0]); Lon0 = dms2d(FaultCoords[1])
            #  Lat1 = dms2d(FaultCoords[2]); Lon1 = dms2d(FaultCoords[3])

            Lat0 = FaultCoords[0]
            Lon0 = FaultCoords[1]
            Lat1 = FaultCoords[2]
            Lon1 = FaultCoords[3]
            Length, Width = np.shape(z)
            Yf0, Xf0 = find_row_column(
                Lon0, Lat0, lon, lat, lon_step, lat_step)
            Yf1, Xf1 = find_row_column(
                Lon1, Lat1, lon, lat, lon_step, lat_step)

            print('*********************************************')
            print(' Fault Coordinates:')
            print('   --------------------------  ')
            print('    Lat          Lon')
            print(str(Lat0) + ' , ' + str(Lon0))
            print(str(Lat1) + ' , ' + str(Lon1))
            print('   --------------------------  ')
            print('    row          column')
            print(str(Yf0) + ' , ' + str(Xf0))
            print(str(Yf1) + ' , ' + str(Xf1))
            print('*********************************************')
            #mf=float(Yf1-Yf0)/float((Xf1-Xf0))  # slope of the fault line
            #cf=float(Yf0-mf*Xf0)   # intercept of the fault line
            #df0=dist_point_from_line(mf,cf,x0,y0,1,1) #distance of the profile start point from the Fault line
            #df1=dist_point_from_line(mf,cf,x1,y1,1,1) #distance of the profile end point from the Fault line

            #mp=-1./mf  # slope of profile which is perpendicualr to the fault line
            # correcting the end point of the profile to be on a line perpendicular to the Fault
            #x1=int((df0+df1)/np.sqrt(1+mp**2)+x0)
            #y1=int(mp*(x1-x0)+y0)

        except:
            print('*********************************************')
            print('No information about the Fault coordinates!')
            print('*********************************************')

        #############################################################################
        y0, x0, y1, x1 = get_start_end_point(
            Xf0, Yf0, Xf1, Yf1, profile_Length, dx, dy)

        try:
            x0
            y0
            x1
            y1
        except:
            fig = plt.figure()
            ax = fig.add_subplot(111)
            ax.imshow(z)
            try:
                ax.plot([Xf0, Xf1], [Yf0, Yf1], 'k-')
            except:
                print('Fault line is not specified')

            xc = []
            yc = []
            print('please click on start and end point of the desired profile')

            def onclick(event):
                if event.button == 1:
                    print('click')
                    xc.append(int(event.xdata))
                    yc.append(int(event.ydata))
            cid = fig.canvas.mpl_connect('button_press_event', onclick)
            plt.show()
            x0 = xc[0]
            x1 = xc[1]
            y0 = yc[0]
            y1 = yc[1]

        ##############################################################################
        print('******************************************************')
        print('First profile coordinates:')
        print('Start point:  y = '+str(y0) + ',x = ' + str(x0))
        print('End point:   y = ' + str(y1) + '  , x = '+str(x1))
        print('')
        print(str(y0) + ',' + str(x0))
        print(str(y1) + ',' + str(x1))
        print('******************************************************')
        length = int(np.hypot(x1-x0, y1-y0))
        x, y = np.linspace(x0, x1, length), np.linspace(y0, y1, length)
        zi = z[y.astype(np.int), x.astype(np.int)]
        try:
            lat_transect = lat_all[y.astype(np.int), x.astype(np.int)]
            lon_transect = lon_all[y.astype(np.int), x.astype(np.int)]
        except:
            lat_transect = 'Nan'
            lon_transect = 'Nan'

        # zi=get_transect(z,x0,y0,x1,y1)

        try:
            dx = float(h5file[k[0]].attrs['X_STEP'])*6375000.0*np.pi/180.0
            dy = float(h5file[k[0]].attrs['Y_STEP'])*6375000.0*np.pi/180.0
            DX = (x-x0)*dx
            DY = (y-y0)*dy
            D = np.hypot(DX, DY)
            print('geo coordinate:')
            print('profile length = ' + str(D[-1]/1000.0) + ' km')
            #df0_km=dist_point_from_line(mf,cf,x0,y0,dx,dy)
        except:
            dx = float(h5file[k[0]].attrs['RANGE_PIXEL_SIZE'])
            dy = float(h5file[k[0]].attrs['AZIMUTH_PIXEL_SIZE'])
            DX = (x-x0)*dx
            DY = (y-y0)*dy
            D = np.hypot(DX, DY)
            print('radar coordinate:')
            print('profile length = ' + str(D[-1]/1000.0) + ' km')
            #df0_km=dist_point_from_line(mf,cf,x0,y0,dx,dy)

        try:
            mf, cf = line(Xf0, Yf0, Xf1, Yf1)
            df0_km = dist_point_from_line(mf, cf, x0, y0, dx, dy)
        except:
            print('Fault line is not specified')

        transect = np.zeros([len(D), ntrans])
        transect[:, 0] = zi
        XX0 = []
        XX1 = []
        YY0 = []
        YY1 = []
        XX0.append(x0)
        XX1.append(x1)
        YY0.append(y0)
        YY1.append(y1)

        if ntrans > 1:

            m = float(y1-y0)/float(x1-x0)
            c = float(y0-m*x0)
            m1 = -1.0/m
            try:
                dp
            except:
                dp = 1.0
            if lat_transect == 'Nan':
                for i in range(1, ntrans):

                    X0 = i*dp/np.sqrt(1+m1**2)+x0
                    Y0 = m1*(X0-x0)+y0
                    X1 = i*dp/np.sqrt(1+m1**2)+x1
                    Y1 = m1*(X1-x1)+y1
                    zi = get_transect(z, X0, Y0, X1, Y1)
                    transect[:, i] = zi
                    XX0.append(X0)
                    XX1.append(X1)
                    YY0.append(Y0)
                    YY1.append(Y1)
            else:
                transect_lat = np.zeros([len(D), ntrans])
                transect_lat[:, 0] = lat_transect
                transect_lon = np.zeros([len(D), ntrans])
                transect_lon[:, 0] = lon_transect

                for i in range(1, ntrans):

                    X0 = i*dp/np.sqrt(1+m1**2)+x0
                    Y0 = m1*(X0-x0)+y0
                    X1 = i*dp/np.sqrt(1+m1**2)+x1
                    Y1 = m1*(X1-x1)+y1
                    zi = get_transect(z, X0, Y0, X1, Y1)
                    lat_transect = get_transect(lat_all, X0, Y0, X1, Y1)
                    lon_transect = get_transect(lon_all, X0, Y0, X1, Y1)
                    transect[:, i] = zi
                    transect_lat[:, i] = lat_transect
                    transect_lon[:, i] = lon_transect
                    XX0.append(X0)
                    XX1.append(X1)
                    YY0.append(Y0)
                    YY1.append(Y1)

        #############################################
        try:
            m_prof_edge, c_prof_edge = line(XX0[0], YY0[0], XX0[-1], YY0[-1])
        except:
            print('Plotting one profile')

        ###############################################################################
        if flip_profile == 'yes':
            transect = np.flipud(transect)
            try:
                df0_km = np.max(D)-df0_km
            except:
                print('')

        print('******************************************************')
        try:
            gpsFile
        except:
            gpsFile = 'Nogps'
        print('GPS velocity file:')
        print(gpsFile)
        print('*******************************************************')
        if os.path.isfile(gpsFile):
            insarData = z
            del z
            Stations, Lat, Lon, Ve, Se, Vn, Sn = readGPSfile(gpsFile, gps_source)
            idxRef = Stations.index(refStation)
            Length, Width = np.shape(insarData)
            lat, lon, lat_step, lon_step = get_lat_lon(h5file)
            IDYref, IDXref = find_row_column(
                Lon[idxRef], Lat[idxRef], lon, lat, lon_step, lat_step)
            if (not np.isnan(IDYref)) and (not np.isnan(IDXref)):
                print('referencing InSAR data to the GPS station at : ' +
                      str(IDYref) + ' , ' + str(IDXref))
                if not np.isnan(insarData[IDYref][IDXref]):
                    transect = transect - insarData[IDYref][IDXref]
                    insarData = insarData - insarData[IDYref][IDXref]

                else:
                    print("""
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

      WARNING: nan value for InSAR data at the reference pixel!
               reference station should be a pixel with valid value in InSAR data.

               please select another GPS station as the reference station.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                   """)
                    sys.exit(1)
            else:
                print('WARNING:')
                print('Reference GPS station is out of the area covered by InSAR data')
                print('please select another GPS station as the reference station.')
                sys.exit(1)

            try:
                stationsList
            except:
                stationsList = Stations

            # theta=23.0*np.pi/180.0
            if os.path.isfile(incidence_file):
                print('Using exact look angle for each pixel')
                h5file_theta = h5py.File(incidence_file, 'r')
                dset = h5file_theta['mask'].get('mask')
                theta = dset[0:dset.shape[0], 0:dset.shape[1]]
                theta = theta*np.pi/180.0
            else:
                print('Using average look angle')
                theta = np.ones(np.shape(insarData))*23.0*np.pi/180.0

            heading = 193.0*np.pi/180.0

            #unitVec=[-np.sin(theta)*np.sin(heading),-np.cos(heading)*np.sin(theta),-np.cos(theta)]
            # -np.cos(theta)]
            unitVec = [np.cos(heading)*np.sin(theta), -
                       np.sin(theta)*np.sin(heading), 0]

            #  [0.0806152480932643, 0.34918300221540616, -0.93358042649720174]
            # print unitVec
            # unitVec=[0.3,-0.09,0.9]
            # unitVec=[-0.3,0.09,-0.9]
            # unitVec=[-0.3,0.09,0]

            # print '*******************************************'
            # print 'unit vector to project GPS to InSAR LOS:'
            # print unitVec
            # print '*******************************************'
            # gpsLOS_ref=unitVec[0]*Ve[idxRef]+unitVec[1]*Vn[idxRef]#+unitVec[2]*Vu[idxRef]

            gpsLOS_ref = gps_to_LOS(
                Ve[idxRef], Vn[idxRef], theta[IDYref, IDXref], heading)
            print('%%%%%%^^^^^^^%%%%%%%%')
            print(gpsLOS_ref/1000.0)
            # insarData=insarData -gpsLOS_ref/1000.0
            # transect = transect -gpsLOS_ref/1000.0

            GPS = []
            GPS_station = []
            GPSx = []
            GPSy = []
            GPS_lat = []
            GPS_lon = []
            for st in stationsList:
                try:
                    idx = Stations.index(st)

                    #gpsLOS = unitVec[0]*Ve[idx]+unitVec[1]*Vn[idx]#+unitVec[2]*Vu[idx]

                    #gpsLOS = gps_to_LOS(Ve[idx],Vn[idx],theta[idx],heading)
                    #gpsLOS=gpsLOS-gpsLOS_ref

                    IDY, IDX = find_row_column(
                        Lon[idx], Lat[idx], lon, lat, lon_step, lat_step)
                    print(theta[IDY, IDX])
                    gpsLOS = gps_to_LOS(
                        Ve[idx], Vn[idx], theta[IDY, IDX], heading)
                    #gpsLOS = gpsLOS-gpsLOS_ref

                    if which_gps == 'all':
                        if theta[IDY, IDX] != 0.0:
                            GPS.append(gpsLOS-gpsLOS_ref)
                            GPS_station.append(st)
                            GPSx.append(IDX)
                            GPSy.append(IDY)
                            GPS_lat.append(Lat[idx])
                            GPS_lon.append(Lon[idx])
                    elif not np.isnan(insarData[IDY][IDX]):
                        if theta[IDY, IDX] != 0.0:
                            GPS.append(gpsLOS-gpsLOS_ref)
                            GPS_station.append(st)
                            GPSx.append(IDX)
                            GPSy.append(IDY)
                            GPS_lat.append(Lat[idx])
                            GPS_lon.append(Lon[idx])
                except:
                    NoInSAR = 'yes'

            DistGPS = []
            GPS_in_bound = []
            GPS_in_bound_st = []
            GPSxx = []
            GPSyy = []
            for i in range(len(GPS_station)):
                gx = GPSx[i]
                gy = GPSy[i]

                if which_gps in ['all', 'insar']:
                    check_result = 'True'
                else:
                    check_result = check_st_in_box(
                        gx, gy, x0, y0, x1, y1, X0, Y0, X1, Y1)

                if check_result == 'True':
                    check_st_in_box2(gx, gy, x0, y0, x1, y1, X0, Y0, X1, Y1)
                    GPS_in_bound_st.append(GPS_station[i])
                    GPS_in_bound.append(GPS[i])
                    GPSxx.append(GPSx[i])
                    GPSyy.append(GPSy[i])
                    # gy=y0+1
                    # gx=x0+1
                    # gxp,gyp=get_intersect(m,c,gx,gy)
                    # Dx=dx*(gx-gxp);Dy=dy*(gy-gyp)
                    # print gxp
                    # print gyp
                    # distance of GPS station from the first profile line
                    dg = dist_point_from_line(m, c, gx, gy, 1, 1)
                    # DistGPS.append(np.hypot(Dx,Dy))
                    # X0=dg/np.sqrt(1+m1**2)+x0
                    # Y0=m1*(X0-x0)+y0
                    # DistGPS.append(np.hypot(dx*(gx-X0), dy*(gy-Y0)))

                    DistGPS.append(dist_point_from_line(
                        m_prof_edge, c_prof_edge, GPSx[i], GPSy[i], dx, dy))

            print('****************************************************')
            print('GPS stations in the profile area:')
            print(GPS_in_bound_st)
            print('****************************************************')
            GPS_in_bound = np.array(GPS_in_bound)
            DistGPS = np.array(DistGPS)
            #axes[1].plot(DistGPS/1000.0, -1*GPS_in_bound/1000, 'bo')

        if gpsFile == 'Nogps':

            insarData = z
            GPSxx = []
            GPSyy = []
            GPSx = []
            GPSy = []
            GPS = []
            XX0[0] = x0
            XX1[0] = x1
            YY0[0] = y0
            YY1[0] = y1

        print('****************')
        print('flip up-down')
        print(flip_updown)

        if flip_updown == 'yes' and gpsFile != 'Nogps':
            print('Flipping up-down')
            transect = -1*transect
            GPS_in_bound = -1*GPS_in_bound
        elif flip_updown == 'yes':
            print('Flipping up-down')
            transect = -1*transect

        if flip_profile == 'yes' and gpsFile != 'Nogps':

            GPS = np.flipud(GPS)
            GPS_in_bound = np.flipud(GPS_in_bound)
            DistGPS = np.flipud(max(D)-DistGPS)

        fig, axes = plt.subplots(nrows=2)
        axes[0].imshow(insarData)
        for i in range(ntrans):
            axes[0].plot([XX0[i], XX1[i]], [YY0[i], YY1[i]], 'r-')

        axes[0].plot(GPSx, GPSy, 'b^')
        axes[0].plot(GPSxx, GPSyy, 'k^')
        if gpsFile != 'Nogps':
            axes[0].plot(IDXref, IDYref, 'r^')
        axes[0].axis('image')
        axes[1].plot(D/1000.0, transect, 'ko', ms=1)

        avgInSAR = np.array(nanmean(transect, axis=1))
        stdInSAR = np.array(nanstd(transect, axis=1))

        # std=np.std(transect,1)
        # axes[1].plot(D/1000.0, avgInSAR, 'r-')
        try:
            axes[1].plot(DistGPS/1000.0, -1*GPS_in_bound/1000, 'b^', ms=10)
        except:
            print('')
        # pl.fill_between(x, y-error, y+error,alpha=0.6, facecolor='0.20')
        # print transect
        #############################################################################

        fig2, axes2 = plt.subplots(nrows=1)
        axes2.imshow(insarData)
        # for i in range(ntrans):
        axes2.plot([XX0[0], XX1[0]], [YY0[0], YY1[0]], 'k-')
        axes2.plot([XX0[-1], XX1[-1]], [YY0[-1], YY1[-1]], 'k-')
        axes2.plot([XX0[0], XX0[-1]], [YY0[0], YY0[-1]], 'k-')
        axes2.plot([XX1[0], XX1[-1]], [YY1[0], YY1[-1]], 'k-')

        try:
            axes2.plot([Xf0, Xf1], [Yf0, Yf1], 'k-')
        except:
            FaultLine = 'None'

        axes2.plot(GPSx, GPSy, 'b^')
        axes2.plot(GPSxx, GPSyy, 'k^')
        if gpsFile != 'Nogps':
            axes2.plot(IDXref, IDYref, 'r^')
        axes2.axis('image')

        figName = 'transect_area_'+str(Np)+'.png'
        print('writing '+figName)
        plt.savefig(figName)

        #############################################################################
        fig = plt.figure()
        fig.set_size_inches(10, 4)
        ax = plt.Axes(fig, [0., 0., 1., 1.], )
        ax = fig.add_subplot(111)
        if display_InSAR in ['on', 'On', 'ON']:
            ax.plot(D/1000.0, transect*1000, 'o',
                    ms=1, mfc='Black', linewidth='0')


        ############################################################################
        # save the profile data:
        if save_to_mat in ['ON', 'on', 'On']:
            import scipy.io as sio
            matFile = 'transect'+str(Np)+'.mat'
            dataset = {}
            dataset['datavec'] = transect
            try:
                dataset['lat'] = transect_lat
                dataset['lon'] = transect_lon
            except:
                dataset['lat'] = 'Nan'
                dataset['lon'] = 'Nan'
            dataset['Unit'] = 'm'
            dataset['Distance_along_profile'] = D
            print('*****************************************')
            print('')
            print('writing transect to >>> '+matFile)
            sio.savemat(matFile, {'dataset': dataset})
            print('')
            print('*****************************************')

        #############################################################################
        if display_Standard_deviation in ['on', 'On', 'ON']:

            for i in np.arange(0.0, 1.01, 0.01):
                # ,color='#DCDCDC')#'LightGrey')
                ax.plot(D/1000.0, (avgInSAR-i*stdInSAR) *
                        1000, '-', color='#DCDCDC', alpha=0.5)
            for i in np.arange(0.0, 1.01, 0.01):
                ax.plot(D/1000.0, (avgInSAR+i*stdInSAR)*1000, '-',
                        color='#DCDCDC', alpha=0.5)  # 'LightGrey')
        #############################################################################
        if display_Average in ['on', 'On', 'ON']:
            ax.plot(D/1000.0, avgInSAR*1000, 'r-')
        ###########
        try:
            ax.plot(DistGPS/1000.0, -1*GPS_in_bound, '^', ms=10, mfc='Cyan')
        except:
            print('')
        ax.set_ylabel('LOS velocity [mm/yr]', fontsize=26)
        ax.set_xlabel('Distance along profile [km]', fontsize=26)

        ###################################################################
        # lower and higher bounds for diplaying the profile

        try:
            lbound
            hbound
        except:
            lbound = np.nanmin(transect)*1000
            hbound = np.nanmax(transect)*1000

        ###################################################################
        # To plot the Fault location on the profile
        ax.plot([df0_km/1000.0, df0_km/1000.0], [lbound, hbound],
                '--', color='black', linewidth='2')

        ###################################################################

        try:
            ax.set_ylim(lbound, hbound)
        except:
            ylim = 'no'

        try:
            ax.set_xlim(x_lbound, x_hbound)
        except:
            xlim = 'no'


        ##########
        # Temporary To plot DEM
        # majorLocator = MultipleLocator(5)
        # ax.yaxis.set_major_locator(majorLocator)
        # minorLocator   = MultipleLocator(1)
        # ax.yaxis.set_minor_locator(minorLocator)

        # plt.tick_params(which='major', length=15,width=2)
        # plt.tick_params(which='minor', length=6,width=2)

        # try:
        #    for tick in ax.xaxis.get_major_ticks():
        #             tick.label.set_fontsize(26)
        #    for tick in ax.yaxis.get_major_ticks():
        #             tick.label.set_fontsize(26)
        #
        #    plt.tick_params(which='major', length=15,width=2)
        #    plt.tick_params(which='minor', length=6,width=2)
        # except:
        #    print 'couldn not fix the ticks! '

        figName = 'transect_'+str(Np)+'.png'
        print('writing '+figName)
        plt.savefig(figName)
        print('')
        print('________________________________')
        # plt.show()


#############################################################################
if __name__ == '__main__':
    main(sys.argv[1:])



#!/usr/bin/env python3
############################################################
# Program is part of MintPy                                #
# Copyright (c) 2013, Zhang Yunjun, Heresh Fattahi         #
# Author: Zhang Yunjun, Antonio Valentino, Aug 2016        #
############################################################


import os
import sys
import time

from mintpy.utils.arg_utils import create_argument_parser

#################################  Usage  ####################################
EXAMPLE = """example:
  temporal_average.py ./inputs/ifgramStack.h5 -d unwrapPhase -o avgPhaseVelocity.h5
  temporal_average.py ./inputs/ifgramStack.h5 -d coherence   -o avgSpatialCoh.h5
"""

def create_parser(subparsers=None):
    synopsis = 'Calculate temporal average (stacking) of multi-temporal datasets'
    epilog = EXAMPLE
    name = __name__.split('.')[-1]
    parser = create_argument_parser(
        name, synopsis=synopsis, description=synopsis, epilog=epilog, subparsers=subparsers)

    parser.add_argument('file', type=str, help='input file with multi-temporal datasets')
    parser.add_argument('-d', '--ds', '--dataset', dest='datasetName', default='coherence',
                        help='dataset name to be averaged, for file with multiple dataset family,\n'+
                        'e.g. ifgramStack.h5\n' +
                        'default: coherence')
    parser.add_argument('-o', '--outfile', help='output file name')
    parser.add_argument('--update', dest='update_mode', action='store_true',
                        help='Enable update checking for --nonzero option.')
    return parser


def cmd_line_parse(iargs=None):
    """Command line parser."""
    # parse
    parser = create_parser()
    inps = parser.parse_args(args=iargs)

    # default:
    if not inps.outfile:
        inps.outfile = auto_output_filename(inps)

    return inps


##############################  Sub Function  ################################
def auto_output_filename(inps):
    """Default output file name"""
    from mintpy.utils import readfile

    ftype = readfile.read_attribute(inps.file)['FILE_TYPE']
    fdir = os.path.dirname(inps.file)
    fbase = os.path.basename(inps.file)

    if ftype == 'ifgramStack':
        if inps.datasetName == 'coherence':
            inps.outfile = 'avgSpatialCoh.h5'
        elif 'unwrapPhase' in inps.datasetName:
            inps.outfile = 'avgPhaseVelocity.h5'
        else:
            inps.outfile = f'avg{inps.datasetName}.h5'

    elif ftype == 'timeseries':
        suffix = os.path.splitext(fbase)[0].split('timeseries')[1]
        inps.outfile = f'avgDisp{suffix}.h5'

    else:
        inps.outfile = f'avg{fbase}'

    inps.outfile = os.path.join(fdir, inps.outfile)
    print(f'output file: {inps.outfile}')

    return inps.outfile


def run_or_skip(inps):
    import h5py

    print('-'*50)
    print('update mode: ON')
    flag = 'skip'

    # check output file vs input dataset
    if not os.path.isfile(inps.outfile):
        flag = 'run'
        print(f'1) output file {inps.outfile} NOT exist.')
    else:
        print(f'1) output file {inps.outfile} already exists.')
        with h5py.File(inps.file, 'r') as f:
            atr = f[inps.datasetName].attrs
            ti = float(atr.get('MODIFICATION_TIME', os.path.getmtime(inps.file)))
        to = os.path.getmtime(inps.outfile)
        if ti > to:
            flag = 'run'
            print(f'2) output file is NOT newer than input dataset: {inps.datasetName}.')
        else:
            print(f'2) output file is newer than input dataset: {inps.datasetName}.')

    # result
    print(f'run or skip: {flag}.')
    return flag


#############################  Main Function  ################################
def main(iargs=None):
    # parse
    inps = cmd_line_parse(iargs)

    # import
    from mintpy.utils import utils1 as ut

    # run or skip
    if inps.update_mode and run_or_skip(inps) == 'skip':
        return

    # run
    start_time = time.time()
    ut.temporal_average(
        inps.file,
        datasetName=inps.datasetName,
        outFile=inps.outfile,
    )

    # used time
    m, s = divmod(time.time()-start_time, 60)
    print(f'time used: {m:02.0f} mins {s:02.1f} secs\n')

    return


##############################################################################
if __name__ == '__main__':
    main(sys.argv[1:])



############################################################
# Program is part of MintPy                                #
# Copyright (c) 2013, Zhang Yunjun, Heresh Fattahi         #
# Author: Zhang Yunjun, Heresh Fattahi, Apr 2020           #
############################################################
# 2020-07: Talib Oliver-Cabrera, add UAVSAR support w/in stripmapStack
# 2020-10: Cunren Liang, add alosStack support
# 2022-06: Yujie Zheng, add standard processing from isce2
# Group contents:
#     metadata
#     geometry
#     baseline
#     multilook
#     miscellaneous
# Recommend import:
#   from mintpy.utils import isce_utils


import datetime
import glob
import logging
import os
import shelve
import time

import numpy as np
from scipy import ndimage

from mintpy.constants import EARTH_RADIUS, SPEED_OF_LIGHT
from mintpy.objects import sensor
from mintpy.utils import (
    attribute as attr,
    ptime,
    readfile,
    utils1 as ut,
    writefile,
)

# suppress matplotlib DEBUG message
mpl_logger = logging.getLogger('matplotlib')
mpl_logger.setLevel(logging.WARNING)



def get_processor(meta_file):
    """
    Get the name of ISCE processor (imaging mode)
    """
    meta_dir = os.path.dirname(meta_file)
    tops_meta_file = os.path.join(meta_dir, 'IW*.xml')
    stripmap_meta_files = [os.path.join(meta_dir, i) for i in ['data.db', 'data.dat', 'data']]
    alosStack_meta_frame_files = glob.glob(os.path.join(meta_dir, 'f1_*', '*.frame.xml'))

    processor = None
    if len(glob.glob(tops_meta_file)) > 0:
        # topsStack
        processor = 'tops'

    elif any(os.path.isfile(i) for i in stripmap_meta_files):
        # stripmapStack
        processor = 'stripmap'

    elif alosStack_meta_frame_files != []:
        # alosStack
        processor = 'alosStack'

    elif meta_file.endswith('.xml'):
        # stripmapApp
        processor = 'stripmap'

    else:
        raise ValueError(f'Un-recognized ISCE processor for metadata file: {meta_file}')
    return processor


#####################################  metadata  #######################################
def load_product(xml_name):
    """Load the product using Product Manager."""
    import isce
    from iscesys.Component.ProductManager import ProductManager
    pm = ProductManager()
    pm.configure()
    return pm.loadProduct(xml_name)


def extract_isce_metadata(meta_file, geom_dir=None, rsc_file=None, update_mode=True):
    """Extract metadata from ISCE stack products
    Parameters: meta_file : str, path of metadata file, reference/IW1.xml or referenceShelve/data.dat
                geom_dir  : str, path of geometry directory.
                rsc_file  : str, output file name of ROIPAC format rsc file. None for not write to disk.
    Returns:    meta      : dict
                frame     : object, isceobj.Scene.Frame.Frame / isceobj.Scene.Burst.Burst
    """
    # check existing rsc_file
    if update_mode and ut.run_or_skip(rsc_file, in_file=meta_file, readable=False) == 'skip':
        return readfile.read_roipac_rsc(rsc_file), None

    # 1. read/extract metadata from XML / shelve file
    processor = get_processor(meta_file)
    if processor == 'tops':
        print('extract metadata from ISCE/topsStack xml file:', meta_file)
        meta, frame = extract_tops_metadata(meta_file)

    elif processor == 'alosStack':
        print('extract metadata from ISCE/alosStack xml file:', meta_file)
        meta, frame = extract_alosStack_metadata(meta_file, geom_dir)

    else:
        print('extract metadata from ISCE/stripmapStack shelve file:', meta_file)
        meta, frame = extract_stripmap_metadata(meta_file)

    # 2. extract metadata from geometry file
    if geom_dir:
        if processor != 'alosStack':
            meta = extract_geometry_metadata(geom_dir, meta)

    # 3. common metadata
    meta['PROCESSOR'] = 'isce'
    if 'ANTENNA_SIDE' not in meta.keys():
        meta['ANTENNA_SIDE'] = '-1'

    # convert all value to string format
    for key, value in meta.items():
        meta[key] = str(value)

    # write to .rsc file
    meta = readfile.standardize_metadata(meta)
    if rsc_file:
        print('writing ', rsc_file)
        writefile.write_roipac_rsc(meta, rsc_file)
    return meta, frame


def extract_tops_metadata(xml_file):
    """Read metadata from xml file for Sentinel-1/TOPS
    Parameters: xml_file : str, path of the .xml file, i.e. reference/IW1.xml
    Returns:    meta     : dict, metadata
                burst    : isceobj.Sensor.TOPS.BurstSLC.BurstSLC object
    """
    import isce
    import isceobj
    from isceobj.Planet.Planet import Planet

    obj = load_product(xml_file)
    burst = obj.bursts[0]
    burstEnd = obj.bursts[-1]

    meta = {}
    meta['prf']             = burst.prf
    meta['startUTC']        = burst.burstStartUTC
    meta['stopUTC']         = burstEnd.burstStopUTC
    meta['radarWavelength'] = burst.radarWavelength
    meta['startingRange']   = burst.startingRange
    meta['passDirection']   = burst.passDirection
    meta['polarization']    = burst.polarization
    meta['trackNumber']     = burst.trackNumber
    meta['orbitNumber']     = burst.orbitNumber

    try:
        meta['PLATFORM'] = sensor.standardize_sensor_name(obj.spacecraftName)
    except:
        if os.path.basename(xml_file).startswith('IW'):
            meta['PLATFORM'] = 'sen'

    time_seconds = (burst.sensingMid.hour * 3600.0 +
                    burst.sensingMid.minute * 60.0 +
                    burst.sensingMid.second)
    meta['CENTER_LINE_UTC'] = time_seconds

    orbit = burst.orbit
    peg = orbit.interpolateOrbit(burst.sensingMid, method='hermite')

    # Sentinel-1 TOPS pixel spacing
    Vs = np.linalg.norm(peg.getVelocity())   #satellite speed
    meta['satelliteSpeed'] = Vs
    meta['azimuthPixelSize'] = Vs * burst.azimuthTimeInterval
    meta['rangePixelSize'] = burst.rangePixelSize

    # Sentinel-1 TOPS spatial resolution
    iw_str = 'IW2'
    if os.path.basename(xml_file).startswith('IW'):
        iw_str = os.path.splitext(os.path.basename(xml_file))[0]
    meta['azimuthResolution'] = sensor.SENSOR_DICT['sen'][iw_str]['azimuth_resolution']
    meta['rangeResolution']   = sensor.SENSOR_DICT['sen'][iw_str]['range_resolution']

    elp = Planet(pname='Earth').ellipsoid
    llh = elp.xyz_to_llh(peg.getPosition())
    elp.setSCH(llh[0], llh[1], orbit.getENUHeading(burst.sensingMid))
    meta['HEADING'] = orbit.getENUHeading(burst.sensingMid)
    meta['earthRadius'] = elp.pegRadCur
    meta['altitude'] = llh[2]

    # for Sentinel-1
    meta['beam_mode'] = 'IW'
    meta['swathNumber'] = burst.swathNumber
    # 1. multipel subswaths
    xml_files = glob.glob(os.path.join(os.path.dirname(xml_file), 'IW*.xml'))
    if len(xml_files) > 1:
        swath_num = [load_product(fname).bursts[0].swathNumber for fname in xml_files]
        meta['swathNumber'] = ''.join(str(i) for i in sorted(swath_num))

    # 2. calculate ASF frame number for Sentinel-1
    meta['firstFrameNumber'] = int(0.2 * (burst.burstStartUTC   - obj.ascendingNodeTime).total_seconds())
    meta['lastFrameNumber']  = int(0.2 * (burstEnd.burstStopUTC - obj.ascendingNodeTime).total_seconds())
    return meta, burst


def extract_stripmap_metadata(meta_file):
    """Read metadata from shelve file for StripMap stack from ISCE
    Parameters: meta_file : str, path of the shelve file, i.e. referenceShelve/data.dat
    Returns:    meta      : dict, metadata
                frame     : isceobj.Scene.Frame.Frame object
    """
    import isce
    import isceobj
    from isceobj.Planet.Planet import Planet

    if os.path.basename(meta_file).startswith('data'):
        # shelve file from stripmapStack
        # referenceShelve/data     for uavsar
        # referenceShelve/data.dat for all the others
        fbase = os.path.splitext(meta_file)[0]
        with shelve.open(fbase, flag='r') as mdb:
            frame = mdb['frame']

    elif meta_file.endswith(".xml"):   #XML file from stripmapApp
        frame = load_product(meta_file)

    else:
        raise ValueError(f'un-recognized isce/stripmap metadata file: {meta_file}')

    meta = {}
    meta['prf']             = frame.PRF
    meta['startUTC']        = frame.sensingStart
    meta['stopUTC']         = frame.sensingStop
    meta['radarWavelength'] = frame.radarWavelegth
    meta['startingRange']   = frame.startingRange
    meta['trackNumber']     = frame.trackNumber
    meta['orbitNumber']     = frame.orbitNumber
    meta['PLATFORM'] = sensor.standardize_sensor_name(frame.platform.getSpacecraftName())
    meta['polarization'] = str(frame.polarization).replace('/', '')
    if meta['polarization'].startswith("b'"):
        meta['polarization'] = meta['polarization'][2:4]

    time_seconds = (frame.sensingMid.hour * 3600.0 +
                    frame.sensingMid.minute * 60.0 +
                    frame.sensingMid.second)
    meta['CENTER_LINE_UTC'] = time_seconds

    orbit = frame.orbit
    peg = orbit.interpolateOrbit(frame.sensingMid, method='hermite')

    Vs = np.linalg.norm(peg.getVelocity())  #satellite speed
    meta['satelliteSpeed'] = Vs
    meta['azimuthResolution'] = frame.platform.antennaLength / 2.0
    meta['azimuthPixelSize'] = Vs / frame.PRF

    frame.getInstrument()
    rgBandwidth = frame.instrument.pulseLength * frame.instrument.chirpSlope
    meta['rangeResolution'] = abs(SPEED_OF_LIGHT / (2.0 * rgBandwidth))
    meta['rangePixelSize'] = frame.instrument.rangePixelSize

    elp = Planet(pname='Earth').ellipsoid
    llh = elp.xyz_to_llh(peg.getPosition())
    elp.setSCH(llh[0], llh[1], orbit.getENUHeading(frame.sensingMid))
    meta['HEADING'] = orbit.getENUHeading(frame.sensingMid)
    meta['earthRadius'] = elp.pegRadCur
    meta['altitude'] = llh[2]

    # for StripMap
    meta['beam_mode'] = 'SM'
    return meta, frame


def extract_alosStack_metadata(meta_file, geom_dir):
    """Read metadata for ISCE/alosStack from the following files:

    pairs/*-*/
        {date1}.track.xml
        f1_{frame1}/{date1}.frame.xml
        f2_{frame2}/{date1}.frame.xml
        ...

    Parameters: meta_file : str, path of the track xml file, i.e. pairs/*-*/150408/track.xml
                geom_dir  : str, path of the geometry directory, i.e. dates_resampled/150408/insar
    Returns:    meta      : dict, metadata
                track     : isceobj.Sensor.MultiMode.Track.Track object
    """

    import isce
    import isceobj
    from isceobj.Planet.Planet import Planet

    track = load_track(os.path.dirname(meta_file), dateStr=os.path.basename(meta_file).strip('.track.xml'))
    rlooks, alooks, width, length = extract_image_size_alosStack(geom_dir)
    spotlightModes, stripmapModes, scansarNominalModes, scansarWideModes, scansarModes = alos2_acquisition_modes()

    meta = {}
    meta['prf']             = track.prf
    meta['startUTC']        = track.sensingStart + datetime.timedelta(seconds=(alooks-1.0)/2.0*track.azimuthLineInterval)
    meta['stopUTC']         = meta['startUTC'] + datetime.timedelta(seconds=(length-1)*alooks*track.azimuthLineInterval)
    meta['radarWavelength'] = track.radarWavelength
    meta['startingRange']   = track.startingRange + (rlooks-1.0)/2.0*track.rangePixelSize
    meta['passDirection']   = track.passDirection.upper()
    meta['polarization']    = track.frames[0].swaths[0].polarization
    #meta['trackNumber']     = track.trackNumber
    #meta['orbitNumber']     = track.orbitNumber

    meta['PLATFORM'] = sensor.standardize_sensor_name('alos2')

    sensingSec = (meta['stopUTC'] - meta['startUTC']).total_seconds()
    sensingMid = meta['startUTC'] + datetime.timedelta(seconds=sensingSec/2.0)
    time_seconds = (sensingMid.hour * 3600.0 +
                    sensingMid.minute * 60.0 +
                    sensingMid.second)
    meta['CENTER_LINE_UTC'] = time_seconds

    peg = track.orbit.interpolateOrbit(sensingMid, method='hermite')
    Vs = np.linalg.norm(peg.getVelocity())
    meta['satelliteSpeed'] = Vs
    meta['azimuthPixelSize'] = Vs * track.azimuthLineInterval
    meta['rangePixelSize'] = track.rangePixelSize

    azBandwidth = track.prf * 0.8
    if track.operationMode in scansarNominalModes:
        azBandwidth /= 5.0
    if track.operationMode in scansarWideModes:
        azBandwidth /= 7.0
    #use a mean burst synchronizatino here
    if track.operationMode in scansarModes:
        azBandwidth *= 0.85

    meta['azimuthResolution'] = Vs * (1.0/azBandwidth)
    meta['rangeResolution']   = 0.5 * SPEED_OF_LIGHT * (1.0/track.frames[0].swaths[0].rangeBandwidth)

    elp = Planet(pname='Earth').ellipsoid
    llh = elp.xyz_to_llh(peg.getPosition())
    elp.setSCH(llh[0], llh[1], track.orbit.getENUHeading(sensingMid))
    meta['HEADING'] = track.orbit.getENUHeading(sensingMid)
    meta['earthRadius'] = elp.pegRadCur
    meta['altitude'] = llh[2]

    meta['beam_mode'] = track.operationMode
    meta['swathNumber'] = ''.join(str(swath.swathNumber) for swath in track.frames[0].swaths)

    meta['firstFrameNumber'] = track.frames[0].frameNumber
    meta['lastFrameNumber']  = track.frames[-1].frameNumber

    meta['ALOOKS'] = alooks
    meta['RLOOKS'] = rlooks

    # NCORRLOOKS for coherence calibration
    rgfact = float(meta['rangeResolution']) / float(meta['rangePixelSize'])
    azfact = float(meta['azimuthResolution']) / float(meta['azimuthPixelSize'])
    meta['NCORRLOOKS'] = meta['RLOOKS'] * meta['ALOOKS'] / (rgfact * azfact)

    # update pixel_size for multilooked data
    meta['rangePixelSize'] *= meta['RLOOKS']
    meta['azimuthPixelSize'] *= meta['ALOOKS']

    # LAT/LON_REF1/2/3/4
    edge = 3
    lat_file = glob.glob(os.path.join(geom_dir, f'*_{rlooks}rlks_{alooks}alks.lat'))[0]
    img = isceobj.createImage()
    img.load(lat_file+'.xml')
    width = img.width
    length = img.length
    data = np.memmap(lat_file, dtype='float64', mode='r', shape=(length, width))
    meta['LAT_REF1'] = str(data[ 0+edge,  0+edge])
    meta['LAT_REF2'] = str(data[ 0+edge, -1-edge])
    meta['LAT_REF3'] = str(data[-1-edge,  0+edge])
    meta['LAT_REF4'] = str(data[-1-edge, -1-edge])

    lon_file = glob.glob(os.path.join(geom_dir, f'*_{rlooks}rlks_{alooks}alks.lon'))[0]
    data = np.memmap(lon_file, dtype='float64', mode='r', shape=(length, width))
    meta['LON_REF1'] = str(data[ 0+edge,  0+edge])
    meta['LON_REF2'] = str(data[ 0+edge, -1-edge])
    meta['LON_REF3'] = str(data[-1-edge,  0+edge])
    meta['LON_REF4'] = str(data[-1-edge, -1-edge])

    los_file = glob.glob(os.path.join(geom_dir, f'*_{rlooks}rlks_{alooks}alks.los'))[0]
    data = np.memmap(los_file, dtype='float32', mode='r', shape=(length*2, width))[0:length*2:2, :]
    inc_angle = data[int(length/2), int(width/2)]
    meta['CENTER_INCIDENCE_ANGLE'] = str(inc_angle)

    pointingDirection = {'right': -1, 'left' :1}
    meta['ANTENNA_SIDE'] = str(pointingDirection[track.pointingDirection])

    return meta, track


def alos2_acquisition_modes():
    '''
    return ALOS-2 acquisition mode
    '''

    spotlightModes = ['SBS']
    stripmapModes = ['UBS', 'UBD', 'HBS', 'HBD', 'HBQ', 'FBS', 'FBD', 'FBQ']
    scansarNominalModes = ['WBS', 'WBD', 'WWS', 'WWD']
    scansarWideModes = ['VBS', 'VBD']
    scansarModes = ['WBS', 'WBD', 'WWS', 'WWD', 'VBS', 'VBD']

    return (spotlightModes, stripmapModes, scansarNominalModes, scansarWideModes, scansarModes)


def extract_image_size_alosStack(geom_dir):
    import isce
    import isceobj

    # grab the number of looks in azimuth / range direction
    lats = glob.glob(os.path.join(geom_dir, '*_*rlks_*alks.lat'))
    rlooks = max(int(os.path.splitext(os.path.basename(x))[0].split('_')[1].strip('rlks')) for x in lats)
    alooks = max(int(os.path.splitext(os.path.basename(x))[0].split('_')[2].strip('alks')) for x in lats)

    # grab the number of rows / coluns
    lat = glob.glob(os.path.join(geom_dir, f'*_{rlooks}rlks_{alooks}alks.lat'))[0]
    img = isceobj.createImage()
    img.load(lat+'.xml')
    width = img.width
    length = img.length

    return (rlooks, alooks, width, length)


def load_track(trackDir, dateStr):
    '''Load the track using Product Manager.

    Parameters: trackDir - str, directory of the *.track.xml file
                dateStr  - str, date in YYMMDD format
    Returns:    track    - isceobj.Sensor.MultiMode.Track.Track object
    '''

    # read *.track.xml file
    track = load_product(os.path.join(trackDir, f'{dateStr}.track.xml'))

    # read *.frame.xml files
    track.frames = []
    fnames = sorted(glob.glob(os.path.join(trackDir, f'f*_*/{dateStr}.frame.xml')))
    for fname in fnames:
        track.frames.append(load_product(fname))

    return track



#####################################  geometry  #######################################
def extract_multilook_number(geom_dir, meta=dict(), fext_list=['.rdr','.geo','.rdr.full','.geo.full']):
    for fbase in ['hgt','lat','lon','los']:
        fbase = os.path.join(geom_dir, fbase)
        for fext in fext_list:
            fnames = glob.glob(fbase+fext)
            if len(fnames) > 0:
                break

        if len(fnames) > 0:
            fullXmlFile = f'{fnames[0]}.full.xml'
            if os.path.isfile(fullXmlFile):
                fullXmlDict = readfile.read_isce_xml(fullXmlFile)
                xmlDict = readfile.read_attribute(fnames[0])
                meta['ALOOKS'] = int(int(fullXmlDict['LENGTH']) / int(xmlDict['LENGTH']))
                meta['RLOOKS'] = int(int(fullXmlDict['WIDTH']) / int(xmlDict['WIDTH']))
                break

    # default value
    for key in ['ALOOKS', 'RLOOKS']:
        if key not in meta:
            meta[key] = 1

    # NCORRLOOKS for coherence calibration
    rgfact = float(meta['rangeResolution']) / float(meta['rangePixelSize'])
    azfact = float(meta['azimuthResolution']) / float(meta['azimuthPixelSize'])
    meta['NCORRLOOKS'] = meta['RLOOKS'] * meta['ALOOKS'] / (rgfact * azfact)
    return meta


def extract_geometry_metadata(geom_dir, meta=dict(), box=None, fbase_list=['hgt','lat','lon','los'],
                              fext_list=['.rdr','.geo','.rdr.full','.geo.full']):
    """Extract / update metadata from geometry files

    extract LAT_REF1/2/3/4 from lat*
    extract LON_REF1/2/3/4 from lon*
    extract HEADING from los (azimuth angle)

    extract A/RLOOKS by comparing hgt.xml and hgt.full.xml file
    update azimuthPixelSize / rangePixelSize based on A/RLOOKS

    extract LENGTH/WIDTH from the first geom file
    update corresponding metadata if box is not None
    """

    def get_nonzero_row_number(data, buffer=2):
        """Find the first and last row number of rows without zero value
        for multiple swaths data
        """
        if np.all(data):
            r0, r1 = 0 + buffer, -1 - buffer
        else:
            row_flag = np.sum(data != 0., axis=1) == data.shape[1]
            row_idx = np.where(row_flag)[0]
            r0, r1 = row_idx[0] + buffer, row_idx[-1] - buffer
        return r0, r1

    # grab existing files
    geom_dir = os.path.abspath(geom_dir)
    for fext in fext_list:
        geom_files = [os.path.join(geom_dir, fbase+fext) for fbase in fbase_list]
        geom_files = [i for i in geom_files if os.path.isfile(i)]
        if len(geom_files) > 0:
            break

    # printout message
    if len(geom_files) == 0:
        msg = 'WARNING: No geometry files found with the following pattern!'
        msg += f'\n    file basenme: {fbase_list}'
        msg += f'\n    file extension: {fext_list}'
        print(msg)
        return meta

    print(f'extract metadata from geometry files: {[os.path.basename(i) for i in geom_files]}')

    # get A/RLOOKS
    meta = extract_multilook_number(geom_dir, meta, fext_list=fext_list)

    # update pixel_size for multilooked data
    meta['rangePixelSize'] *= meta['RLOOKS']
    meta['azimuthPixelSize'] *= meta['ALOOKS']

    # get LENGTH/WIDTH
    atr = readfile.read_attribute(geom_files[0])
    meta['LENGTH'] = atr['LENGTH']
    meta['WIDTH'] = atr['WIDTH']

    # update due to subset
    if box:
        meta = attr.update_attribute4subset(meta, box)

    # get LAT/LON_REF1/2/3/4 into metadata
    for geom_file in geom_files:
        if 'lat' in os.path.basename(geom_file):
            data = readfile.read(geom_file, box=box)[0]
            r0, r1 = get_nonzero_row_number(data)
            meta['LAT_REF1'] = str(data[r0, 0])
            meta['LAT_REF2'] = str(data[r0, -1])
            meta['LAT_REF3'] = str(data[r1, 0])
            meta['LAT_REF4'] = str(data[r1, -1])

        if 'lon' in os.path.basename(geom_file):
            data = readfile.read(geom_file, box=box)[0]
            r0, r1 = get_nonzero_row_number(data)
            meta['LON_REF1'] = str(data[r0, 0])
            meta['LON_REF2'] = str(data[r0, -1])
            meta['LON_REF3'] = str(data[r1, 0])
            meta['LON_REF4'] = str(data[r1, -1])

        if 'los' in os.path.basename(geom_file):
            # CENTER_INCIDENCE_ANGLE
            data = readfile.read(geom_file, datasetName='inc', box=box)[0]
            data[data == 0.] = np.nan
            inc_angle = data[int(data.shape[0]/2), int(data.shape[1]/2)]
            meta['CENTER_INCIDENCE_ANGLE'] = str(inc_angle)
    return meta



#####################################  baseline  #######################################
def read_tops_baseline(baseline_file):
    """Read baseline file generated by ISCE/topsStack processor.
    Example:
    baselines/20141213_20160418/20141213_20160418.txt:
        swath: IW1
        Bperp (average): 62.62863491739495
        Bpar (average): -29.435602419751426
        swath: IW2
        Bperp (average): 60.562020649374034
        Bpar (average): -34.56105358031081
    """
    bperps = []
    with open(baseline_file) as f:
        for line in f:
            c = line.split(":")
            if c[0] == "Bperp (average)":
                bperps.append(float(c[1]))
    bperp_top = np.mean(bperps)
    bperp_bottom = np.mean(bperps)
    return [bperp_top, bperp_bottom]


def read_stripmap_baseline(baseline_file):
    """Read baseline file generated by ISCE/stripmapStack processor.

    Example:
    baselines/20200111_20200125.txt
        PERP_BASELINE_BOTTOM 173.97914535263297
        PERP_BASELINE_TOP 174.05612879066618
    """
    fDict = readfile.read_template(baseline_file, delimiter=' ')
    bperp_top = float(fDict['PERP_BASELINE_TOP'])
    bperp_bottom = float(fDict['PERP_BASELINE_BOTTOM'])
    return [bperp_top, bperp_bottom]


def read_alosStack_baseline(baseline_file):
    '''read baseline file generated by alosStack
    '''
    bDict = {}
    with open(baseline_file) as f:
        lines = [line for line in f if line.strip() != '']
        for x in lines[2:]:
            blist = x.split()
            #to fit into the format of other processors, all alos satellites are after 2000
            blist[0] = '20' + blist[0]
            blist[1] = '20' + blist[1]
            bDict[blist[1]] = [float(blist[3]), float(blist[3])]
        bDict[blist[0]] = [0, 0]

    return bDict, blist[0]


def read_baseline_timeseries(baseline_dir, processor='tops', ref_date=None):
    """Read bperp time-series from files in baselines directory
    Parameters: baseline_dir : str, path to the baselines directory
                processor    : str, tops     for Sentinel-1/TOPS
                                    stripmap for StripMap data
                ref_date     : str, reference date in (YY)YYMMDD
    Returns:    bDict : dict, in the following format:
                    {'20141213': [0.0, 0.0],
                     '20141225': [104.6, 110.1],
                     ...
                    }
    """

    # grab all existed baseline files
    print(f'read perp baseline time-series from {baseline_dir}')
    if processor == 'tops':
        bFiles = sorted(glob.glob(os.path.join(baseline_dir, '*/*.txt')))
    elif processor == 'stripmap':
        bFiles = sorted(glob.glob(os.path.join(baseline_dir, '*.txt')))
    elif processor == 'alosStack':
        # all baselines are in baseline_center.txt
        bFiles = glob.glob(os.path.join(baseline_dir, 'baseline_center.txt'))
    else:
        raise ValueError(f'Un-recognized ISCE stack processor: {processor}')

    if len(bFiles) == 0:
        print(f'WARNING: no baseline text file found in dir {os.path.abspath(baseline_dir)}')
        return None

    if processor in ['tops', 'stripmap']:
        # ignore files with different date1
        # when re-run with different reference date
        date1s = [os.path.basename(i).split('_')[0] for i in bFiles]
        date1c = ut.most_common(date1s)
        bFiles = [i for i in bFiles if os.path.basename(i).split('_')[0] == date1c]

        # ignore empty files
        bFiles = [i for i in bFiles if os.path.getsize(i) > 0]

        # read files into dict
        bDict = {}
        for bFile in bFiles:
            dates = os.path.basename(bFile).split('.txt')[0].split('_')
            if processor == 'tops':
                bDict[dates[1]] = read_tops_baseline(bFile)
            else:
                bDict[dates[1]] = read_stripmap_baseline(bFile)
        bDict[dates[0]] = [0, 0]
        ref_date0 = dates[0]

    elif processor == 'alosStack':
        bDict, ref_date0 = read_alosStack_baseline(bFiles[0])

    else:
        raise ValueError(f'Un-recognized ISCE stack processor: {processor}')

    # change reference date
    if ref_date is not None and ref_date != ref_date0:
        ref_date = ptime.yyyymmdd(ref_date)
        print(f'change reference date to {ref_date}')
        ref_bperp = bDict[ref_date]

        for key in bDict.keys():
            bDict[key][0] -= ref_bperp[0]
            bDict[key][1] -= ref_bperp[1]

    return bDict



#####################################  multilook  #######################################

def multilook_number2resolution(meta_file, az_looks, rg_looks):
    # get full resolution info
    az_pixel_size, az_spacing, rg_pixel_size, rg_spacing = get_full_resolution(meta_file)

    # print out message
    print(f'Azimuth     pixel size : {az_pixel_size:.1f}')
    print(f'Azimuth ground spacing : {az_spacing:.1f}')
    print(f'Azimuth ground spacing : {az_spacing*az_looks:.1f} after multilooking by {az_looks}')

    print(f'Range       pixel size : {rg_pixel_size:.1f}')
    print(f'Range   ground spacing : {rg_spacing:.1f}')
    print(f'Range   ground spacing : {rg_spacing*rg_looks:.1f} after multilooking by {rg_looks}')
    return


def resolution2multilook_number(meta_file, resolution):
    """
    Calculate multilook number for InSAR processing given a disired output resolution on the ground

    Parameters: meta_file   : str, path of ISCE metadata file, i.e. IW1.xml, data.dat
                resolution  : float, target output resolution on the ground in meters
    Returns:    az/rg_looks : int, number of looks in azimuth / range direction
    """
    # get full resolution info
    az_pixel_size, az_spacing, rg_pixel_size, rg_spacing = get_full_resolution(meta_file)

    # calculate number of looks
    # 1. adjust the final resolution in one direction closest to the input resolution
    az_looks = resolution / az_spacing
    rg_looks = resolution / rg_spacing
    az_round_frac = abs(az_looks - np.rint(az_looks))
    rg_round_frac = abs(rg_looks - np.rint(rg_looks))
    if az_round_frac < rg_round_frac:
        resolution = np.rint(az_looks) * az_spacing
    else:
        resolution = np.rint(rg_looks) * rg_spacing

    # 2. calculate the multilook number based on the adjusted resolution
    az_looks = np.rint(resolution / az_spacing).astype(int)
    rg_looks = np.rint(resolution / rg_spacing).astype(int)

    # print out message
    print(f'Azimuth     pixel size : {az_pixel_size:.1f}')
    print(f'Azimuth ground spacing : {az_spacing:.1f}')
    print(f'Azimuth ground spacing : {az_spacing*az_looks:.1f} after multilooking by {az_looks}')

    print(f'Range       pixel size : {rg_pixel_size:.1f}')
    print(f'Range   ground spacing : {rg_spacing:.1f}')
    print(f'Range   ground spacing : {rg_spacing*rg_looks:.1f} after multilooking by {rg_looks}')

    return az_looks, rg_looks


def get_full_resolution(meta_file):
    """
    Grab the full resolution in terms of pixel_size and ground spacing
    """
    # check metadata file extension: only ISCE format is supported.
    fext = os.path.splitext(meta_file)[1]
    if fext not in ['.xml', '.dat']:
        raise ValueError(f'input ISCE metadata file extension "{fext}" not in [.xml, .dat]')

    # get middle sub-swath xml file for Sentinel-1 data
    if meta_file.endswith('.xml'):
        meta_files = glob.glob(meta_file)
        mid_idx = int(len(meta_files) / 2)
        meta_file = meta_files[mid_idx]

    # extract metadata
    meta, frame = extract_isce_metadata(meta_file, update_mode=False)
    meta['WIDTH'] = frame.numberOfSamples

    # calculate the full azimuth/range ground resolution
    az_pixel_size = float(meta['AZIMUTH_PIXEL_SIZE'])  #azimuth pixel size on the orbit
    rg_pixel_size = float(meta['RANGE_PIXEL_SIZE'])    #range   pixel size in LOS direction

    height = float(meta['HEIGHT'])
    inc_angle = ut.incidence_angle(meta, dimension=0)

    az_spacing = az_pixel_size * EARTH_RADIUS / (EARTH_RADIUS + height)  #azimuth pixel size on the ground
    rg_spacing = rg_pixel_size / np.sin(inc_angle / 180. * np.pi)        #range   pixel size on the ground

    return az_pixel_size, az_spacing, rg_pixel_size, rg_spacing



#####################################  miscellaneous  #######################################

def get_IPF(proj_dir, ts_file):
    """Grab the IPF version number of each sub-swatch for Sentinel-1 time-series

    Parameters: proj_dir    - str, path of the project directory
                              E.g.: ~/data/AtacamaSenDT149
                ts_file     - str, path of HDF5 file for time-series
    Returns:    date_list   - list of str, dates in YYYYMMDD format
                IFP_IW1/2/3 - list of str, IFP version number
    """
    from mintpy.objects import timeseries

    s_dir = os.path.join(proj_dir, 'secondarys')
    m_dir = os.path.join(proj_dir, 'reference')

    # date list
    date_list = timeseries(ts_file).get_date_list()
    num_date = len(date_list)
    # reference date
    m_date = [i for i in date_list if not os.path.isdir(os.path.join(s_dir, i))][0]

    # grab IPF numver
    IPF_IW1, IPF_IW2, IPF_IW3 = [], [], []
    prog_bar = ptime.progressBar(maxValue=num_date)
    for i in range(num_date):
        date_str = date_list[i]

        # get xml_dir
        if date_str == m_date:
            xml_dir = m_dir
        else:
            xml_dir = os.path.join(s_dir, date_str)

        # grab IPF version number
        for j, IPF_IW in enumerate([IPF_IW1, IPF_IW2, IPF_IW3]):
            xml_file = os.path.join(xml_dir, f'IW{j+1}.xml')
            IPFv = load_product(xml_file).processingSoftwareVersion
            IPF_IW.append(f'{float(IPFv):.02f}')

        prog_bar.update(i+1, suffix=f'{date_str} IW1/2/3')
    prog_bar.close()
    return date_list, IPF_IW1, IPF_IW2, IPF_IW3


def get_sensing_datetime_list(proj_dir, date_list=None):
    """Get the sensing datetime objects from ISCE stack results.
    It assumes the default directory structure from topsStack, as below:
    /proj_dir
        /reference/IW*.xml
        /secondarys
            /20150521/IW*.xml
            /20150614/IW*.xml
            ...
            /20210113/IW*.xml

    Parameters: proj_dir     - str, path to the root directory of stack processing
    Returns:    sensingMid   - list of datetime.datetime.obj
                sensingStart - list of datetime.datetime.obj
                sensingStop  - list of datetime.datetime.obj
    """
    # determine xml file basename
    ref_fname = glob.glob(os.path.join(proj_dir, 'reference', 'IW*.xml'))[0]
    fbase = os.path.basename(ref_fname)

    # get xml files for all acquisitions
    sec_fnames = sorted(glob.glob(os.path.join(proj_dir, 'secondarys', '*', fbase)))
    fnames = [ref_fname] + sec_fnames
    num_file = len(fnames)

    # loop to read file one by one
    sensingStart = []
    sensingStop = []
    for i, fname in enumerate(fnames):
        print(f'[{i+1}/{num_file}] read {fname}')
        obj = load_product(fname)
        sensingStart.append(obj.bursts[0].sensingStart)
        sensingStop.append(obj.bursts[-1].sensingStop)

    sensingStart = sorted(sensingStart)
    sensingStop  = sorted(sensingStop)

    # sensingStart/Stop --> sensingMid
    sensingMid = [i + (j - i)/2 for i, j in zip(sensingStart, sensingStop)]

    # round to the nearest second
    print('round sensingStart/Stop/Mid to the nearest second.')
    sensingStart = [ptime.round_seconds(i) for i in sensingStart]
    sensingStop  = [ptime.round_seconds(i) for i in sensingStop]
    sensingMid   = [ptime.round_seconds(i) for i in sensingMid]

    if date_list is not None:
        date_str_format = ptime.get_date_str_format(date_list[0])
        date_list_out = [i.strftime(date_str_format) for i in sensingMid]

        # check possible missing dates
        dates_missing = [i for i in date_list if i not in date_list_out]
        if dates_missing:
            raise ValueError(f'The following dates are missing:\n{dates_missing}')

        # prune dates not-needed
        flag = np.array([i in date_list for i in date_list_out], dtype=np.bool_)
        if np.sum(flag) > 0:
            sensingMid    = np.array(sensingMid)[flag].tolist()
            sensingStart  = np.array(sensingStart)[flag].tolist()
            sensingStop   = np.array(sensingStop)[flag].tolist()
            dates_removed = np.array(date_list_out)[~flag].tolist()
            print(f'The following dates are not needed and removed:\n{dates_removed}')

    return sensingMid, sensingStart, sensingStop



############################## Standard Processing ###########################################

def gaussian_kernel(sx, sy, sig_x, sig_y):
    '''Generate a guassian kernal (with all elements sum to 1).

    Parameters: sx/y    - int, dimensions of kernal
                sig_x/y - float, standard deviation of the guassian distribution
    '''
    # ensure sx/y are odd number
    sx += 1 if np.mod(sx, 2) == 0 else 0
    sy += 1 if np.mod(sy, 2) == 0 else 0

    x, y = np.meshgrid(np.arange(sx), np.arange(sy))
    x += 1
    y += 1

    xc = (sx + 1) / 2
    yc = (sy + 1) / 2
    fx = ((x-xc)**2.) / (2.*sig_x**2.)
    fy = ((y-yc)**2.) / (2.*sig_y**2.)

    k = np.exp(-1.0 * (fx+fy))
    a = 1./np.sum(k)
    k = a * k

    return k


def convolve(data, kernel):
    '''Convolve / filter the complex data based on the given kernel.

    Parameters: data   - 2D np.ndarray in complex
                kernel - 2D np.ndarray in float, convolution kernel
    '''
    real = ndimage.convolve(data.real, kernel, mode='constant', cval=0.0)
    imag = ndimage.convolve(data.imag, kernel, mode='constant', cval=0.0)
    return real + 1J * imag


def estimate_coherence(intfile, corfile):
    '''Estimate the spatial coherence (phase sigma) of the wrapped interferogram.

    Parameters: intfile - str, path to the *.int file
                corfile - str, path to the output correlation file
    '''
    import isce
    import isceobj
    from mroipac.icu.Icu import Icu

    # create filt interferogram file object
    filtImage = isceobj.createIntImage()
    filtImage.load(intfile + '.xml')
    filtImage.setAccessMode('read')
    filtImage.createImage()

    # create phase sigma correlation file object
    phsigImage = isceobj.createImage()
    phsigImage.dataType='FLOAT'
    phsigImage.bands = 1
    phsigImage.setWidth(filtImage.getWidth())
    phsigImage.setFilename(corfile)
    phsigImage.setAccessMode('write')
    phsigImage.createImage()

    # setup Icu() object
    icuObj = Icu(name='sentinel_filter_icu')
    icuObj.configure()
    icuObj.unwrappingFlag = False
    icuObj.useAmplitudeFlag = False
    #icuObj.correlationType = 'NOSLOPE'

    # run
    icuObj.icu(intImage=filtImage, phsigImage=phsigImage)
    phsigImage.renderHdr()

    # close
    filtImage.finalizeImage()
    phsigImage.finalizeImage()

    return


def unwrap_snaphu(int_file, cor_file, unw_file, defo_max=2.0, max_comp=32,
                  init_only=True, init_method='MCF', cost_mode='SMOOTH'):
    '''Unwrap interferograms using SNAPHU via isce2.

    Modified from ISCE-2/topsStack/unwrap.py
    Notes from Piyush:
        SNAPHU is an iterative solver, starting from the initial solution. It can get
            stuck in an infinite loop.
        The initial solution is created using MCF or MST method. The MST initial solution
            typically require lots of iterations and may not be a good starting point.
        DEFO cost mode requires geometry info for the program to interpret the coherence
            correctly and setup costs based on that. DEFO always sounds more theoretical
            to me, but I haven not fully explored it. TOPO cost mode requires spatial baseline.
            SMOOTH cost mode is purely data driven.
        Amplitude of zero is a mask in all cost modes. For TOPO mode, amplitude is used to find
            layover; for SMOOTH mode, only non-zero amplitude matters.

    Default configurations in ISCE-2/topsStack:
        init_only = True
        init_method = 'MCF'
        cost_mode = 'SMOOTH'
    Default configurations in FRInGE:
        init_only = False
        init_method = 'MST'
        cost_mode = 'DEFO'

    Parameters: int_file    - str, path to the wrapped interferogram file
                cor_file    - str, path to the correlation file
                unw_file    - str, path to the output unwrapped interferogram file
                defo_max    - float, maximum number of cycles for the deformation phase
                max_comp    - int, maximum number of connected components
                init_only   - bool, initlize-only mode
                init_method - str, algo used for initialization: MCF, MST
                cost_mode   - str, statistical-cost mode: TOPO, DEFO, SMOOTH, NOSTATCOSTS
    Returns:    unw_file    - str, path to the output unwrapped interferogram file
    '''
    import isce
    from contrib.Snaphu.Snaphu import Snaphu

    start_time = time.time()

    # configurations - atr
    atr = readfile.read_attribute(int_file)
    width = int(atr['WIDTH'])
    length = int(atr['LENGTH'])
    altitude = float(atr['HEIGHT'])
    earth_radius = float(atr['EARTH_RADIUS'])
    wavelength = float(atr['WAVELENGTH'])
    rg_looks = int(atr['RLOOKS'])
    az_looks = int(atr['ALOOKS'])
    corr_looks = float(atr.get('NCORRLOOKS', rg_looks * az_looks / 1.94))

    ## setup SNAPHU
    # https://web.stanford.edu/group/radar/softwareandlinks/sw/snaphu/snaphu.conf.full
    # https://github.com/isce-framework/isce2/blob/main/contrib/Snaphu/Snaphu.py
    print('phase unwrapping with SNAPHU ...')
    print(f'SNAPHU cost mode: {cost_mode}')
    print(f'SNAPHU init only: {init_only}')
    print(f'SNAPHU init method: {init_method}')
    print(f'SNAPHU max number of connected components: {max_comp}')

    snp = Snaphu()

    # file IO
    snp.setInput(int_file)
    snp.setOutput(unw_file)
    snp.setCorrfile(cor_file)
    snp.setCorFileFormat('FLOAT_DATA')
    snp.setWidth(width)

    # runtime options
    snp.setCostMode(cost_mode)
    snp.setInitOnly(init_only)
    snp.setInitMethod(init_method)

    # geometry parameters
    # baseline info is not used in deformation mode, but is very important in topography mode
    snp.setAltitude(altitude)
    snp.setEarthRadius(earth_radius)
    snp.setWavelength(wavelength)
    snp.setRangeLooks(rg_looks)
    snp.setAzimuthLooks(az_looks)
    snp.setCorrLooks(corr_looks)

    # deformation mode parameters
    snp.setDefoMaxCycles(defo_max)

    # connected component control
    # grow connectedc components if init_only is True
    # https://github.com/isce-framework/isce2/blob/main/contrib/Snaphu/Snaphu.py#L413
    snp.setMaxComponents(max_comp)

    ## run SNAPHU
    snp.prepare()
    snp.unwrap()
    print('finished SNAPHU running')

    # mask out wired values from SNAPHU
    # based on https://github.com/isce-framework/isce2/pull/326
    flag = np.fromfile(int_file, dtype=np.complex64).reshape(length, width)
    data = np.memmap(unw_file, dtype='float32', mode='r+', shape=(length*2, width))
    data[0:length*2:2, :][np.nonzero(flag == 0)] = 0
    data[1:length*2:2, :][np.nonzero(flag == 0)] = 0

    ## render metadata
    print(f'write metadata file: {unw_file}.xml')
    atr['FILE_TYPE'] = '.unw'
    atr['DATA_TYPE'] = 'float32'
    atr['INTERLEAVE'] = 'BIL'
    atr['BANDS'] = '2'
    writefile.write_isce_xml(atr, unw_file)

    if snp.dumpConnectedComponents:
        print(f'write metadata file: {unw_file}.conncomp.xml')
        atr['FILE_TYPE'] = '.conncomp'
        atr['DATA_TYPE'] = 'uint8'
        atr['INTERLEAVE'] = 'BIP'
        atr['BANDS'] = '1'
        writefile.write_isce_xml(atr, f'{unw_file}.conncomp')

    # time usage
    m, s = divmod(time.time() - start_time, 60)
    print(f'time used: {m:02.0f} mins {s:02.1f} secs.')

    return unw_file



############################################################
# Program is part of MintPy                                #
# Copyright (c) 2013, Zhang Yunjun, Heresh Fattahi         #
# Author: Zhang Yunjun, Heresh Fattahi, 2013               #
############################################################


import numpy as np

from mintpy.objects import sensor
from mintpy.utils import ptime, readfile, utils as ut, writefile


################################################################################
def get_overlap_lalo(atr_list):
    """Find overlap area in lat/lon of geocoded files based on their metadata.
    Parameters: atr_list - list of dict, attribute dictionary of two input files in geo coord
    Returns:    S/N/W/E  - float, West/East/South/North in deg
    """
    S, N, W, E = None, None, None, None
    for i, atr in enumerate(atr_list):
        Si, Ni, Wi, Ei = ut.four_corners(atr)
        if i == 0:
            S, N, W, E = Si, Ni, Wi, Ei
        else:
            S = max(Si, S)
            N = min(Ni, N)
            W = max(Wi, W)
            E = min(Ei, E)

    return S, N, W, E


def get_design_matrix4east_north_up(los_inc_angle, los_az_angle, obs_direction):
    """Design matrix G to convert multi-track range/azimuth displacement into east/north/up direction.
    Parameters: los_inc_angle - 1D np.ndarray in size of (num_obs,) in float32, LOS incidence angle in degree
                los_az_angle  - 1D np.ndarray in size of (num_obs,) in float32, LOS azimuth   angle in degree
                obs_direction - 1D np.ndarray in size of (num_obs,) in str, observation direction: range or azimuth
    Returns:    G             - 2D np.ndarray in size of (num_obs, 3) in float32, design matrix
    """
    num_obs = los_inc_angle.shape[0]
    G = np.zeros((num_obs, 3), dtype=np.float32)

    for i, (inc_angle, az_angle, obs_dir) in enumerate(zip(los_inc_angle, los_az_angle, obs_direction)):
        # calculate the unit vector
        if obs_dir == 'range':
            # for range offset / InSAR phase [with positive value for motion toward the satellite]
            ve = np.sin(np.deg2rad(inc_angle)) * np.sin(np.deg2rad(az_angle)) * -1
            vn = np.sin(np.deg2rad(inc_angle)) * np.cos(np.deg2rad(az_angle))
            vu = np.cos(np.deg2rad(inc_angle))

        elif obs_dir == 'azimuth':
            # for azimuth offset [with positive value for motion same as flight]
            ve = np.sin(np.deg2rad(az_angle - 90)) * -1
            vn = np.cos(np.deg2rad(az_angle - 90)) * 1
            vu = 0.

        else:
            raise ValueError(f'un-recognized observation direction: {obs_dir}')

        # fill the design matrix
        G[i, :] = [ve, vn, vu]

    return G


def get_design_matrix4horz_vert(los_inc_angle, los_az_angle, horz_az_angle=-90):
    """Design matrix G to convert asc/desc range displacement into horz/vert direction.
    Only asc + desc -> hz + up is implemented for now.

    Project displacement from LOS to Horizontal and Vertical components:
    Math for 3D:
        dLOS =   dE * sin(inc_angle) * sin(az_angle) * -1
               + dN * sin(inc_angle) * cos(az_angle)
               + dU * cos(inc_angle)
    Math for 2D:
        dLOS =   dH * sin(inc_angle) * cos(az_angle - az)
               + dV * cos(inc_angle)
        with dH_perp = 0.0
    This could be easily modified to support multiple view geometry
        (e.g. two adjcent tracks from asc & desc) to resolve 3D

    Parameters: los_inc_angle - 1D np.ndarray in size of (num_file), LOS incidence angle in degree.
                los_az_angle  - 1D np.ndarray in size of (num_file), LOS azimuth   angle in degree.
                horz_az_angle - float, azimuth angle for the horizontal direction of interest in degree.
                                Measured from the north with anti-clockwise direction as positive.
    Returns:    G             - 2D matrix in size of (num_file, 2)
    """
    num_file = los_inc_angle.shape[0]
    G = np.zeros((num_file, 2), dtype=np.float32)
    for i in range(num_file):
        G[i, 0] = np.sin(np.deg2rad(los_inc_angle[i])) * np.cos(np.deg2rad(los_az_angle[i] - horz_az_angle))
        G[i, 1] = np.cos(np.deg2rad(los_inc_angle[i]))

    return G


def asc_desc2horz_vert(dlos, los_inc_angle, los_az_angle, horz_az_angle=-90, step=20):
    """Decompose asc / desc LOS data into horz / vert data.
    Parameters: dlos          - 3D np.ndarray in size of (num_file, length, width), LOS displacement in meters.
                los_inc_angle - 1/3D np.ndarray in size of (num_file), length, width), LOS incidence angle in degree.
                los_az_angle  - 1/3D np.ndarray in size of (num_file), length, width), LOS azimuth   angle in degree.
                horz_az_angle - float, horizontal azimuth angle of interest in degree.
                step          - int, geometry step size
    Returns:    dhorz         - 2D np.ndarray in size of (length, width), horizontal displacement in meters.
                dvert         - 2D np.ndarray in size of (length, width), vertical   displacement in meters.
    """
    # initiate output
    (num_file, length, width) = dlos.shape
    dhorz = np.zeros((length, width), dtype=np.float32) * np.nan
    dvert = np.zeros((length, width), dtype=np.float32) * np.nan

    # 0D (constant) incidence / azimuth angle --> invert once for all pixels
    if los_inc_angle.ndim == 1:
        G = get_design_matrix4horz_vert(los_inc_angle, los_az_angle, horz_az_angle)
        print('decomposing asc/desc into horz/vert direction ...')
        dhv = np.dot(np.linalg.pinv(G), dlos.reshape(num_file, -1)).astype(np.float32)
        dhorz = dhv[0, :].reshape(length, width)
        dvert = dhv[1, :].reshape(length, width)

    # 2D incidence / azimuth angle --> invert window-by-window
    elif los_inc_angle.ndim == 3:
        num_row = np.ceil(length / step).astype(int)
        num_col = np.ceil(width / step).astype(int)

        print(f'decomposing asc/desc into horz/vert direction in windows of {step}x{step} ...')
        prog_bar = ptime.progressBar(maxValue=num_row)
        for i in range(num_row):
            y0, y1 = step * i, min(step * (i + 1), length)
            for j in range(num_col):
                x0, x1 = step * j, min(step * (j + 1), width)

                # calculate the median geometry for the local window
                med_los_inc_angle = np.nanmedian(los_inc_angle[:, y0:y1, x0:x1], axis=(1,2))
                med_los_az_angle  = np.nanmedian( los_az_angle[:, y0:y1, x0:x1], axis=(1,2))
                if np.all(~np.isnan(med_los_inc_angle)):

                    G = get_design_matrix4horz_vert(med_los_inc_angle, med_los_az_angle, horz_az_angle)
                    dhv = np.dot(np.linalg.pinv(G), dlos[:, y0:y1, x0:x1].reshape(num_file, -1))
                    dhorz[y0:y1, x0:x1] = dhv[0].reshape(y1-y0, x1-x0)
                    dvert[y0:y1, x0:x1] = dhv[1].reshape(y1-y0, x1-x0)

            prog_bar.update(i+1, suffix=f'{i+1}/{num_row}')
        prog_bar.close()

    else:
        raise ValueError(f'un-supported incidence angle matrix dimension ({los_inc_angle.ndim})!')

    return dhorz, dvert


def run_asc_desc2horz_vert(inps):
    """Decompose asc / desc LOS files into horz / vert file(s).
    Parameters: inps         - namespace, input parameters
    Returns:    inps.outfile - str(s) output file(s)
    """

    ## 1. calculate the overlaping area in lat/lon
    atr_list = [readfile.read_attribute(fname, datasetName=inps.ds_name) for fname in inps.file]
    S, N, W, E = get_overlap_lalo(atr_list)
    lat_step = float(atr_list[0]['Y_STEP'])
    lon_step = float(atr_list[0]['X_STEP'])
    length = int(round((S - N) / lat_step))
    width  = int(round((E - W) / lon_step))
    print(f'overlaping area in SNWE: {(S, N, W, E)}')


    ## 2. read LOS data and geometry
    num_file = len(inps.file)
    dlos = np.zeros((num_file, length, width), dtype=np.float32)
    if inps.geom_file:
        los_inc_angle = np.zeros((num_file, length, width), dtype=np.float32)
        los_az_angle  = np.zeros((num_file, length, width), dtype=np.float32)
    else:
        los_inc_angle = np.zeros(num_file, dtype=np.float32)
        los_az_angle  = np.zeros(num_file, dtype=np.float32)

    for i, (atr, fname) in enumerate(zip(atr_list, inps.file)):
        # overlap SNWE --> box to read for each specific file
        coord = ut.coordinate(atr)
        x0 = coord.lalo2yx(W, coord_type='lon')
        y0 = coord.lalo2yx(N, coord_type='lat')
        box = (x0, y0, x0 + width, y0 + length)

        # read data
        dlos[i, :] = readfile.read(fname, box=box, datasetName=inps.ds_name)[0]
        msg = f'{inps.ds_name} ' if inps.ds_name else ''
        print(f'read {msg} from file: {fname}')

        # read geometry
        if inps.geom_file:
            los_inc_angle[i, :] = readfile.read(inps.geom_file[i], box=box, datasetName='incidenceAngle')[0]
            los_az_angle[i, :]  = readfile.read(inps.geom_file[i], box=box, datasetName='azimuthAngle')[0]
            print(f'read 2D LOS incidence / azimuth angles from file: {inps.geom_file[i]}')
        else:
            los_inc_angle[i] = ut.incidence_angle(atr, dimension=0, print_msg=False)
            los_az_angle[i] = ut.heading2azimuth_angle(float(atr['HEADING']))
            print('calculate the constant LOS incidence / azimuth angles from metadata as:')
            print(f'LOS incidence angle: {los_inc_angle[i]:.1f} deg')
            print(f'LOS azimuth   angle: {los_az_angle[i]:.1f} deg')


    ## 3. decompose LOS displacements into horizontal / Vertical displacements
    print('---------------------')
    dhorz, dvert = asc_desc2horz_vert(dlos, los_inc_angle, los_az_angle, inps.horz_az_angle)


    ## 4. write outputs
    print('---------------------')
    # Update attributes
    atr = atr_list[0].copy()
    if inps.ds_name and atr['FILE_TYPE'] in ['ifgramStack', 'timeseries', 'HDFEOS']:
        atr['FILE_TYPE'] = 'displacement'

    atr['WIDTH']  = str(width)
    atr['LENGTH'] = str(length)
    atr['X_STEP'] = str(lon_step)
    atr['Y_STEP'] = str(lat_step)
    atr['X_FIRST'] = str(W)
    atr['Y_FIRST'] = str(N)

    # update REF_X/Y
    ref_lat, ref_lon = float(atr['REF_LAT']), float(atr['REF_LON'])
    [ref_y, ref_x] = ut.coordinate(atr).geo2radar(ref_lat, ref_lon)[0:2]
    atr['REF_Y'] = int(ref_y)
    atr['REF_X'] = int(ref_x)

    # use ref_file for time-series file writing
    ref_file = inps.file[0] if atr_list[0]['FILE_TYPE'] == 'timeseries' else None

    if inps.one_outfile:
        print(f'write asc/desc/horz/vert datasets into {inps.one_outfile}')
        dsDict = {}
        for i, atr_i in enumerate(atr_list):
            # dataset name for LOS data
            track_num = atr_i.get('trackNumber', None)
            proj_name = atr_i.get('PROJECT_NAME', None)
            if proj_name in ['none', 'None', None]:
                proj_name = atr_i.get('FILE_PATH', None)
            proj_name = sensor.project_name2sensor_name(proj_name)[0]

            ds_name = proj_name if proj_name else ''
            ds_name += 'A' if atr_i['ORBIT_DIRECTION'].lower().startswith('asc') else 'D'
            ds_name += f'T{track_num}' if track_num else ''
            ds_name += '_{}'.format(atr_i['DATE12'])

            # assign dataset value
            dsDict[ds_name] = dlos[i]
        dsDict['horizontal'] = dhorz
        dsDict['vertical'] = dvert
        writefile.write(dsDict, out_file=inps.one_outfile, metadata=atr, ref_file=ref_file)

    else:
        print('writing horizontal component to file: '+inps.outfile[0])
        writefile.write(dhorz, out_file=inps.outfile[0], metadata=atr, ref_file=ref_file)
        print('writing vertical   component to file: '+inps.outfile[1])
        writefile.write(dvert, out_file=inps.outfile[1], metadata=atr, ref_file=ref_file)

    return inps.outfile



#!/usr/bin/env python3
############################################################
# Program is part of MintPy                                #
# Copyright (c) 2013, Zhang Yunjun, Heresh Fattahi         #
# Author: Heresh Fattahi, 2013                             #
############################################################


import getopt
import os
import sys

import h5py
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import MultipleLocator


def readGPSfile(gpsFile, gps_source):
    if gps_source in ['cmm4', 'CMM4']:

        gpsData = np.loadtxt(gpsFile, usecols=(1, 2, 3, 4, 5, 6, 7, 8, 9, 10))
        Stations = np.loadtxt(gpsFile, dtype=str, usecols=(0, 1))[:, 0]

        St = []
        Lon = []
        Lat = []
        Ve = []
        Vn = []
        Se = []
        Sn = []
        for i in range(gpsData.shape[0]):
            if 'GPS' in Stations[i]:
                Lat.append(gpsData[i, 0])
                Lon.append(gpsData[i, 1])
                Ve.append(gpsData[i, 2])
                Se.append(gpsData[i, 3])
                Vn.append(gpsData[i, 4])
                Sn.append(gpsData[i, 5])
                St.append(Stations[i])
        Up = np.zeros(St.shape)
        Sup = np.zeros(St.shape)
    elif gps_source == 'mintpy':

        #gpsData = np.loadtxt(gpsFile,usecols = (1,2,3,4,5,6,7,8,9))
        gpsData = np.loadtxt(gpsFile, usecols=(1, 2, 3, 4, 5, 6))
        Stations = np.loadtxt(gpsFile, dtype=str, usecols=(0, 1))[:, 0]

        St = []
        Lon = []
        Lat = []
        Ve = []
        Vn = []
        Se = []
        Sn = []

        for i in range(gpsData.shape[0]):
            Lon.append(gpsData[i, 0]-360)
            Lat.append(gpsData[i, 1])
            Ve.append(gpsData[i, 2])
            Vn.append(gpsData[i, 3])
            Se.append(gpsData[i, 4])
            Sn.append(gpsData[i, 5])
            St.append(Stations[i])
        Up = np.zeros(St.shape)
        Sup = np.zeros(St.shape)
    elif gps_source in ['usgs', 'USGS']:
        gpsData_Hz = np.loadtxt(gpsFile, usecols=(0, 1, 2, 3, 4, 5, 6))
        gpsData_up = np.loadtxt(gpsFile, usecols=(8, 9))
        gpsData = np.hstack((gpsData_Hz, gpsData_up))
        Stations = np.loadtxt(gpsFile, dtype=str, usecols=(7, 8))[:, 0]

        St = []
        Lon = []
        Lat = []
        Ve = []
        Vn = []
        Se = []
        Sn = []
        Up = []
        Sup = []

        for i in range(gpsData.shape[0]):
            Lat.append(gpsData[i, 0])
            Lon.append(gpsData[i, 1])
            Vn.append(gpsData[i, 2])
            Ve.append(gpsData[i, 3])
            Sn.append(gpsData[i, 4])
            Se.append(gpsData[i, 5])
            Up.append(gpsData[i, 6])
            Sup.append(gpsData[i, 7])
            St.append(Stations[i])

    return list(St), Lat, Lon, Ve, Se, Vn, Sn, Up, Sup


def nearest(x, tbase, xstep):
    # """ find nearest neighbour """
    dist = np.sqrt((tbase - x)**2)
    if min(dist) <= np.abs(xstep):
        indx = dist == min(dist)
    else:
        indx = []
    return indx


def find_row_column(Lon, Lat, lon, lat, lon_step, lat_step):
    ################################################
    # finding row and column numbers of the GPS point

    idx = nearest(Lon, lon, lon_step)
    idy = nearest(Lat, lat, lat_step)
    if idx != [] and idy != []:
        IDX = np.where(idx == True)[0][0]
        IDY = np.where(idy == True)[0][0]
    else:
        IDX = np.nan
        IDY = np.nan
    return IDY, IDX


################################################
def usage():
    print("""
   ************************************************************************************************

   Compares InSAR and GPS velocities. Option G can be used to specify the mode of Comparison.
   Out put is InSARvsGPS.png

   Usage:

   insar_vs_gps.py  -v InSARvelocity.h5 -g GPS velocity file -r Name of the reference GPS station -l ststion list to be compared

   -l :  if station list is not specified, all stations in GPS velocity file are compared with InSAR

   -m min value of the x and y axis of the plot
   -M max value of the x and y axis of the plot
   -r refernce GPS station
   -s second velocity map
   -S source of the GPS data: (usgs,cmm4,mintpy)
      see documentation for more infromation
   -I incidence angle (if not given average look angle is used instead)
   -H Heading angle (if not given then the program reads it from the attributes of the velocity file)

   -G GPS components to be used to compare with InSAR: los_3D , los_Hz , los_Up , gps_Up. [default is los_3D]
       los_3D: to project three gps components to LOS
       los_Hz: to project horizontal gps components to LOS
       los_Up: to project only vertical gps component to LOS
       gps_Up: uses vertical GPS to compare with InSAR (InSAR LOS will be projected to Up)

   -A annotate the GPS station name on the plot [yes] or no
   -C annotation font color [default is green]
   -x annotation offset from the point in x direction [default=0]
   -y annotation offset from the point in y direction [default=0]
   -u to plot 1, 2 or 3 sigma uncertainty [default=1]
   -B marker size [default = 15]
   Example:

    insar_vs_gps.py  -v geo_InSAR_velocity.h5  -g gpsVelocity.txt -S usgs  -r BEMT -l 'HNPS,DHLG,SLMS,USGC'
    insar_vs_gps.py  -v geo_InSAR_velocity.h5  -g gpsVelocity.txt -S cmm4 -r BEMT -l 'HNPS,DHLG,SLMS,USGC,ROCH,MONP,SIO3,IVCO,TMAP,BMHL,BILL,OGHS'
    insar_vs_gps.py  -v geo_InSAR_velocity.h5  -g gpsVelocity.txt -S usgs -r BEMT
    insar_vs_gps.py  -v geo_InSAR_velocity.h5  -g gpsVelocity.txt -S usgs -r BEMT -c geo_temporal_coherence.h5 -t 0.95
    insar_vs_gps.py  -v geo_InSAR_velocity.h5  -g gpsVelocity.txt -S usgs -r BEMT -c geo_temporal_coherence.h5 -t 0.95  -l 'HNPS,DHLG,SLMS,USGC'

    insar_vs_gps.py  -v geo_velocity_New_masked.h5 -g usgs_velocities_NAfixed.txt -r BEMT -S usgs -A yes -C green -x 0 -y 0.5 -H 193.0 -I 23.0

   +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

   When user wants to compare two InSAR velocities at GPS stations:

   Example: (to comaper geo_InSAR_velocity.h5 with simulated_velocity.h5 in GPS stations)

    insar_vs_gps.py  -v geo_InSAR_velocity.h5  -g gpsVelocity.txt  -r BEMT -s simulated_velocity.h5

   ************************************************************************************************
    """)


def main(argv):

    annotation = 'yes'
    ann_x = 0
    ann_y = 0
    annotation_Color = 'green'
    disp_velocity = 'yes'
    GPS_InSAR_dif_thr = 1
    gps_comp = 'los_3D'
    uncertainty_fac = 1.0
    MarkerSize = 5
    try:
        opts, args = getopt.getopt(
            argv, "v:r:g:G:l:c:t:m:M:s:S:A:B:C:x:y:I:H:u:")

    except getopt.GetoptError:
        usage()
        sys.exit(1)

    for opt, arg in opts:

        if opt == '-v':
            velocityFile = arg
        elif opt == '-s':
            velocityFile2 = arg
        elif opt == '-g':
            gpsFile = arg
        elif opt == '-r':
            refStation = arg
        elif opt == '-l':
            stationsList = arg.split(',')
        elif opt == '-c':
            coherenceFile = arg
        elif opt == '-t':
            thr = float(arg)
        elif opt == '-m':
            minV = float(arg)
        elif opt == '-M':
            maxV = float(arg)
        elif opt == '-S':
            gps_source = arg
        elif opt == '-A':
            annotation = arg
        elif opt == '-C':
            annotation_Color = arg
        elif opt == '-x':
            ann_x = float(arg)
        elif opt == '-y':
            ann_y = float(arg)
        elif opt == '-I':
            theta = float(arg)
        elif opt == '-H':
            heading = float(arg)
        elif opt == '-G':
            gps_comp = arg
        elif opt == '-u':
            uncertainty_fac = float(arg)
        elif opt == '-B':
            MarkerSize = float(arg)

    try:
        velocityFile
        gpsFile
        refStation
    except:
        usage()
        sys.exit(1)

    try:
        thr
    except:
        thr = 0.9

    h5file = h5py.File(velocityFile, 'r')
    dset = h5file['velocity'].get('velocity')
    insarData = dset[0:dset.shape[0], 0:dset.shape[1]]
    k = list(h5file.keys())

    try:
        h5file2 = h5py.File(velocityFile2, 'r')
        dset2 = h5file2['velocity'].get('velocity')
        insarData2 = dset2[0:dset2.shape[0], 0:dset2.shape[1]]
    except:
        print('')

    ullon = float(h5file[k[0]].attrs['X_FIRST'])
    ullat = float(h5file[k[0]].attrs['Y_FIRST'])
    lon_step = float(h5file[k[0]].attrs['X_STEP'])
    lat_step = float(h5file[k[0]].attrs['Y_STEP'])
    #lon_unit = h5file[k[0]].attrs['Y_UNIT']
    #lat_unit = h5file[k[0]].attrs['X_UNIT']

    Length, Width = np.shape(insarData)

    lllat = ullat+Length*lat_step
    urlon = ullon+Width*lon_step
    lat = np.arange(ullat, lllat, lat_step)
    lon = np.arange(ullon, urlon, lon_step)
    #################################################################################################
    # finding the raw an column of the reference gps station and referencing insar data to this pixel
    Stations, Lat, Lon, Ve, Se, Vn, Sn, Vu, Su = readGPSfile(
        gpsFile, gps_source)
    idxRef = Stations.index(refStation)
    IDYref, IDXref = find_row_column(
        Lon[idxRef], Lat[idxRef], lon, lat, lon_step, lat_step)

    #############################################
    #  Stations, gpsData = redGPSfile(gpsFile)
    #  idxRef=Stations.index(refStation)
    #  Lat,Lon,Vn,Ve,Sn,Se,Corr,Vu,Su = gpsData[idxRef,:]
    #  IDYref,IDXref=find_row_column(Lon,Lat,lon,lat,lon_step,lat_step)
    ###################################################

    if (not np.isnan(IDYref)) and (not np.isnan(IDXref)):
        print('')
        print('-----------------------------------------------------------------------')
        print('referencing InSAR data to the GPS station at : ' +
              str(IDYref) + ' , ' + str(IDXref))
        if not np.isnan(insarData[IDYref][IDXref]):
            insarData = insarData - insarData[IDYref][IDXref]
        else:
            print("""
%%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
       WARNING: nan value for InSAR data at the refernce pixel!
                reference station should be a pixel with valid value in InSAR data.

                please select another GPS station as the reference station.
%%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                  """)
            sys.exit(1)
    else:
        print('WARNING:')
        print('Reference GPS station is out of the area covered by InSAR data')
        print('please select another GPS station as the reference station.')
        sys.exit(1)

    #######################################################################################

    try:
        stationsList
    except:
        stationsList = Stations
        # stationsList.remove(refStation)

    try:
        print('incidence angle = ' + str(theta))
    except:
        print('using look angle from the velocity file. For more precise results input the incidence angle using option -I.')
        look_n = float(h5file['velocity'].attrs['LOOK_REF1'])
        look_f = float(h5file['velocity'].attrs['LOOK_REF2'])
        theta = (look_n+look_f)/2.
        print('incidence angle = ' + str(theta))

    try:
        print('Heading angle = '+str(heading))
    except:
        heading = float(h5file['velocity'].attrs['HEADING'])
        if heading < 0:
            heading = heading+360

    theta = theta*np.pi/180.0
    heading = heading*np.pi/180.0

    if gps_comp in ['los_3D', 'LOS_3D', 'los_3d']:
        unitVec = [np.cos(heading)*np.sin(theta), -np.sin(theta)
                   * np.sin(heading), -np.cos(theta)]
        gps_comp_txt = ' projecting three gps components to LOS'
    elif gps_comp in ['los_Hz', 'LOS_HZ', 'los_hz', 'los_HZ', 'LOS_hz']:
        unitVec = [np.cos(heading)*np.sin(theta), -
                   np.sin(theta)*np.sin(heading), 0]
        gps_comp_txt = ' projecting horizontal gps components to LOS'
    elif gps_comp in ['LOS_UP', 'los_Up', 'los_up', 'LOS_up']:
        unitVec = [0, 0, -np.cos(theta)]
        gps_comp_txt = ' projecting vertical gps components to LOS'
    elif gps_comp in ['gps_up', 'GPS_UP', 'GPS_Up', 'gps_Up']:
        unitVec = [0, 0, 1]
        gps_comp_txt = ' comparing veryical gps with InSAR'
        print('-------------------------')
        print('Projecting InSAR to vertical')
        insarData = -insarData/np.cos(theta)
    print('-------------------------')
    print('unit vector for :' + gps_comp_txt)
    print(unitVec)
    print('-------------------------')
    gpsLOS_ref = (unitVec[0] * Ve[idxRef]
                  + unitVec[1] * Vn[idxRef]
                  + unitVec[2] * Vu[idxRef])
    Sr = ((unitVec[0]**2) * Se[idxRef]**2
          + (unitVec[1]**2) * Sn[idxRef]**2
          + (unitVec[2]**2) * Su[idxRef]**2)**0.5

    print('######################################################################')
    try:
        h5coh = h5py.File(coherenceFile)
        kh5coh = list(h5coh.keys())
        dset = h5coh[kh5coh[0]].get(kh5coh[0])
        Coh = dset[0:dset.shape[0], 0:dset.shape[1]]
    except:
        print('No information about the coherence of the points')

    InSAR = []
    GPS = []
    InSAR1 = []
    GPS1 = []
    InSAR2 = []
    GPS2 = []
    coherence = []
    GPSx = []
    GPSy = []
    GPSx1 = []
    GPSy1 = []
    GPSx2 = []
    GPSy2 = []
    GPS_station = []
    GPS_std = []
    for st in stationsList:

        try:
            idx = Stations.index(st)
            #Lat,Lon,Vn,Ve,Sn,Se,Corr,Vu,Su = gpsData[idx,:]
            gpsLOS = unitVec[0]*Ve[idx]+unitVec[1]*Vn[idx]+unitVec[2]*Vu[idx]
            Sg = ((unitVec[0]**2)*Se[idx]**2+(unitVec[1]**2)
                  * Sn[idx]**2+(unitVec[2]**2)*Su[idx]**2)**0.5
            S = (Sg**2+Sr**2)**0.5
            gpsLOS = gpsLOS-gpsLOS_ref
            IDY, IDX = find_row_column(
                Lon[idx], Lat[idx], lon, lat, lon_step, lat_step)
            insar_velocity = -insarData[IDY][IDX]

            try:
                gpsLOS = insarData2[IDY][IDX]-insarData2[IDYref][IDXref]
                gpsLOS = -1000.0*gpsLOS
            except:
                InSAR_GPS_Copmarison = 'yes'

            if not np.isnan(insarData[IDY][IDX]):
                print('%%%%%%%%%%%%%%%%%%%%')
                print(st)
                print('GPS: ' + str(gpsLOS) + '  +/- '+str(S))
                print('INSAR: ' + str(-insarData[IDY][IDX]*1000.0))
                try:
                    print('Coherence: ' + str(Coh[IDY][IDX]))
                    coherence.append(Coh[IDY][IDX])
                    if Coh[IDY][IDX] > thr:
                        InSAR1.append(-insarData[IDY][IDX]*1000.0)
                        GPS1.append(gpsLOS)
                    else:
                        InSAR2.append(-insarData[IDY][IDX]*1000.0)
                        GPS2.append(gpsLOS)

                except:
                    print('No information about the coherence is available!')

                InSAR.append(-insarData[IDY][IDX]*1000.0)
                GPS.append(gpsLOS)
                GPS_station.append(st)
                GPSx.append(IDX)
                GPSy.append(IDY)
                GPS_std.append(S)
                if np.abs(gpsLOS+insarData[IDY][IDX]*1000.0) < GPS_InSAR_dif_thr:
                    GPSx1.append(IDX)
                    GPSy1.append(IDY)
                else:
                    GPSx2.append(IDX)
                    GPSy2.append(IDY)
        except:
            NoInSAR = 'yes'

    InSAR = np.array(InSAR)
    GPS = np.array(GPS)
    GPS_std = np.array(GPS_std)
    lt = len(InSAR)
    SAD = np.sum(np.abs(InSAR-GPS), 0)/lt
    C1 = np.zeros([2, len(InSAR)])
    C1[0][:] = InSAR
    C1[1][:] = GPS
    Cor = np.corrcoef(C1)[0][1]
    print('++++++++++++++++++++++++++++++++++++++++++++++')
    print('Comparison summary:')
    print('')
    print('AAD (average absolute difference)= '+str(SAD) + ' [mm/yr]')
    print('Correlation = '+str(Cor))
    print('')
    print('++++++++++++++++++++++++++++++++++++++++++++++')
    ###############################################################
    try:
        minV
        maxV
    except:
        minV = np.min([InSAR, GPS])
        maxV = np.max([InSAR, GPS])

    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.errorbar(GPS, InSAR, yerr=0.0, xerr=uncertainty_fac *
                GPS_std, fmt='ko', ms=MarkerSize)
    ax.plot([minV-3, maxV+3], [minV-3, maxV+3], 'k--')
    ax.set_ylabel('InSAR [mm/yr]', fontsize=26)
    ax.set_xlabel('GPS LOS [mm/yr]', fontsize=26)
    ax.set_ylim(minV-3, maxV+3)
    ax.set_xlim(minV-3, maxV+3)

    if annotation in ['yes', 'y', 'Y', 'Yes', 'YES']:
        for i in range(len(GPS)):
            ax.annotate(GPS_station[i], xy=(GPS[i], InSAR[i]), xytext=(
                GPS[i]+ann_x, InSAR[i]+ann_y), color=annotation_Color)

    majorLocator = MultipleLocator(5)
    ax.yaxis.set_major_locator(majorLocator)
    minorLocator = MultipleLocator(1)
    ax.yaxis.set_minor_locator(minorLocator)
    ax.xaxis.set_minor_locator(minorLocator)
    for tick in ax.xaxis.get_major_ticks():
        tick.label.set_fontsize(26)
    for tick in ax.yaxis.get_major_ticks():
        tick.label.set_fontsize(26)
    plt.tick_params(which='major', length=15, width=2)
    plt.tick_params(which='minor', length=6, width=2)

    figName = 'InSARvsGPS_errorbar.png'
    plt.savefig(figName)
    ###############################################################

    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.plot(GPS, InSAR, 'ko', ms=MarkerSize)
    ax.plot([minV-3, maxV+3], [minV-3, maxV+3], 'k--')
    # ax.plot([-10,20],[-10,20],'k--')
    ax.set_ylabel('InSAR [mm/yr]', fontsize=26)
    ax.set_xlabel('GPS LOS [mm/yr]', fontsize=26)
    ax.set_ylim(minV-3, maxV+3)
    ax.set_xlim(minV-3, maxV+3)
    # ax.set_ylim(-10,15)
    # ax.set_xlim(-10,15)
    if annotation in ['yes', 'y', 'Y', 'Yes', 'YES']:
        for i in range(len(GPS)):
            ax.annotate(GPS_station[i], xy=(GPS[i], InSAR[i]), xytext=(
                GPS[i]+ann_x, InSAR[i]+ann_y), color=annotation_Color)

    majorLocator = MultipleLocator(5)
    ax.yaxis.set_major_locator(majorLocator)
    minorLocator = MultipleLocator(1)
    ax.yaxis.set_minor_locator(minorLocator)
    ax.xaxis.set_minor_locator(minorLocator)
    for tick in ax.xaxis.get_major_ticks():
        tick.label.set_fontsize(26)
    for tick in ax.yaxis.get_major_ticks():
        tick.label.set_fontsize(26)
    plt.tick_params(which='major', length=15, width=2)
    plt.tick_params(which='minor', length=6, width=2)

    figName = 'InSARvsGPS.png'
    # plt.savefig(figName,pad_inches=0.0)
    plt.savefig(figName)

    ######################################################

    try:
        Coh
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.errorbar(GPS1, InSAR1, yerr=1.0, xerr=1.0, fmt='o')
        ax.errorbar(GPS2, InSAR2, yerr=1.0, xerr=1.0, fmt='^')
        ax.plot([minV-3, maxV+3], [minV-3, maxV+3], '--')
        ax.set_ylabel('InSAR [mm/yr]', fontsize=26)
        ax.set_xlabel('GPS LOS [mm/yr]', fontsize=26)
        ax.set_ylim(minV-3, maxV+3)
        ax.set_xlim(minV-3, maxV+3)

    except:
        print('')
    plt.show()


if __name__ == '__main__':
    main(sys.argv[1:])



#!/usr/bin/env python3
############################################################
# Program is part of MintPy                                #
# Copyright (c) 2013, Zhang Yunjun, Heresh Fattahi         #
# Author: Antonio Valentino, Marin Govorcin, Aug 2022      #
############################################################


import sys

from mintpy.utils import arg_utils

#########################################################################################################
EXAMPLE = """example:
  ## displacement [event-type inversion]
  # option 1: use velocity file with step estimation from timeseries2velocity.py for co-seismic displacement
  save_kite.py geo/geo_velocity.h5 -d step20210104 -g geo/geo_geometry.h5 -m geo/geo_maskTempCoh.h5 -o dsc

  # option 2: use time-series / ifgramStack file with date1_date2 for the transient displacement:
  save_kite.py geo/geo_timeseries_ERA5_ramp_demErr.h5 -d 20101120_20110220 -g geo/geo_geometry.h5 -m geo/geo_maskTempCoh.h5 -o dsc
  save_kite.py geo/geo_ifgramStack.h5     -d unwrapPhase-20101120_20110220 -g geo/geo_geometry.h5 -m geo/geo_maskTempCoh.h5 -o dsc

  ## velocity [interseismic or tensile dislocation inversion]
  # https://pyrocko.org/beat/docs/current/examples/Rectangular_tensile.html
  save_kite.py geo/geo_velocity.h5 -d velocity -g geo/geo_geometry.h5 -m geo/geo_maskTempCoh.h5 -o dsc

  ## import to kite
  spool outfile_name    % /do quadtree,covariance/aps and then File>Save Scene and it is ready for GROND or BEAT
"""

KITE_URL = 'https://github.com/pyrocko/kite'


def create_parser(subparsers=None):
    synopsis = f'Generate KITE ({KITE_URL}) npz and yaml from MintPy HDF5 file.'
    epilog = EXAMPLE
    name = __name__.split('.')[-1]
    parser = arg_utils.create_argument_parser(
        name, synopsis=synopsis, description=synopsis, epilog=epilog, subparsers=subparsers)

    parser.add_argument('file', type=str, help='file to be converted, in geo coordinate.')
    parser.add_argument('-d', '--dset', '--dataset', dest='dset', type=str, required=True,
                        help='dataset of interest to be converted.\n'+
                             'e.g.: velocity / stepYYYYMMDD for velocity HDF5 file,\n'+
                             '      date12 in YYYYMMDD_YYYYMMDD for time-series HDF5 file,\n'+
                             '      date12 in unwrapPhase-YYYYMMDD_YYYYMMDD for ifgramStack HDF5 file.')
    parser.add_argument('-g', '--geom', dest='geom_file', type=str,
                        help='geometry file for incidence /azimuth angle and height.')
    parser.add_argument('-m', '--mask', dest='mask_file', type=str,
                        help='mask file, or run mask.py to mask the input file beforehand.')
    parser.add_argument('-o', '--output', dest='outfile', type=str,
                        help='output filename')

    parser = arg_utils.add_subset_argument(parser)
    return parser


def cmd_line_parse(iargs=None):
    parser = create_parser()
    inps = parser.parse_args(args=iargs)
    return inps


#########################################################################################################
def main(iargs=None):
    # parse
    inps = cmd_line_parse(iargs)

    # import
    from mintpy.save_kite import save_kite

    # run
    save_kite(inps)


#########################################################################################################
if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python3
# Author: Zhang Yunjun, Nov 2022
"""Test topographic residuel correction due to DEM errors."""


import argparse
import datetime
import math
import sys

import numpy as np
from matplotlib import pyplot as plt

from mintpy.dem_error import estimate_dem_error
from mintpy.utils import ptime, time_func

################################################################################
# truth
delta_z_sim = 50   # meters

# setting: SAR geometry
range_dist = 800e3  # meters
inc_angle = 34      # degree
max_pbase = 500     # meters

# setting: time / acquisition
revisit_time = datetime.timedelta(days=24)
start_date = datetime.datetime(2018, 1, 1)
num_date = 50
ref_ind = 10

# setting: dem_error.py
phase_velocity = False
cond = 1e-8


################################################################################
EXAMPLE = """example:
  $MINTPY_HOME/tests/dem_error.py
  $MINTPY_HOME/tests/dem_error.py --plot
"""

def cmd_line_parse(iargs=None):
    # create parser
    parser = argparse.ArgumentParser(description='Test dem_error.py',
                                     formatter_class=argparse.RawTextHelpFormatter,
                                     epilog=EXAMPLE)
    parser.add_argument('--plot', dest='plot', action='store_true', help='Plot testing results.')

    # parsing
    inps = parser.parse_args(args=iargs)

    return inps


################################################################################
def sim_pbase_and_topo_residual(num_date, delta_z, ref_ind=0,
                                max_pbase=500, inc_angle=34, range_dist=800e3):

    # sim perp baseline TS
    np.random.seed(12138)
    pbase = np.random.rand(num_date) * max_pbase
    pbase -= pbase[ref_ind]

    # sim topo residual phase
    G_geom = pbase / (range_dist * np.sin(np.deg2rad(inc_angle)))
    ts_topo_res = np.dot(G_geom, delta_z_sim)

    return pbase, ts_topo_res


def plot_result(date_list, ts_sim, ts_obs=None, ts_cor=None, model=None):

    dt_list = ptime.date_list2vector(date_list)[0]

    _, ax = plt.subplots(nrows=1, ncols=1, figsize=(12, 4))
    ax.plot(dt_list, ts_sim*100, '--', color='C0', lw=3, label='simulation')
    if ts_obs is not None:
        ax.plot(dt_list, ts_obs*100, '.', color='C1', lw=3, label='observation')
    if ts_cor is not None:
        ax.plot(dt_list, ts_cor*100, 'o',  color='C2', ms=12, fillstyle='none', label='corrected')

    # plot step function
    if model is not None and 'stepDate' in model.keys():
        for step_date in model['stepDate']:
            step_dt = datetime.datetime.strptime(step_date, '%Y%m%d')
            ax.axvline(step_dt, ls='--', color='k')

    # axis format
    ax.set_xlabel('Time')
    ax.set_ylabel('Displacement [cm]')
    ax.legend()
    plt.show()


################################################################################
def test_dem_error_with_linear_defo(date_list, tbase, rel_tol=0.05, plot=False):
    print('Test 1: simple time-series with linear displacement.')

    # setting
    vel_sim = 0.05  # m/yr
    model = {'polynomial' : 1}

    # simulate displacement time-series
    ts_sim = vel_sim * tbase
    ts_sim -= ts_sim[ref_ind]

    # add topo residual
    pbase, ts_topo_res = sim_pbase_and_topo_residual(
        num_date,
        delta_z_sim,
        ref_ind,
        max_pbase=max_pbase,
        inc_angle=inc_angle,
        range_dist=range_dist,
    )

    # observed time-series: displacement + topo res
    ts_obs = ts_sim + ts_topo_res

    # estimate DEM error
    G_defo = time_func.get_design_matrix4time_func(date_list, model)
    G_geom = np.reshape(pbase / (range_dist * np.sin(np.deg2rad(inc_angle))), (-1,1))
    G = np.hstack((G_geom, G_defo))

    delta_z_est, ts_cor = estimate_dem_error(
        ts_obs, G, tbase,
        phase_velocity=phase_velocity,
        cond=cond,
    )[:2]

    # plot
    if plot:
        plot_result(date_list, ts_sim, ts_obs, ts_cor, model)

    # validate
    print(f'Specified DEM error: {delta_z_sim:.2f} m')
    print(f'Estimated DEM error: {delta_z_est[0]:.2f} m')
    assert math.isclose(delta_z_sim, delta_z_est, rel_tol=rel_tol)
    print('Pass.')


def test_dem_error_with_complex_defo(date_list, tbase, rel_tol=0.05, plot=False):
    print('Test 2: complex time-series with highly non-linear displacement.')

    # setting
    model = {'polynomial' : 2, 'stepDate' : ['20190818', '20200812']}

    # simulate displacement time-series
    # run the following to re-generate:
    #   from mintpy.simulation import simulation as sim
    #   ts_sim = sim.sim_variable_timeseries(50)
    ts_sim = np.array([
         0.        , -0.00049281, -0.00098563, -0.00147844, -0.00197125,
        -0.00246407, -0.00295688, -0.00344969, -0.0039425 , -0.00443532,
         0.06961907,  0.08992067,  0.10159181,  0.10972946,  0.11593095,
         0.12090778,  0.12503949,  0.12855262,  0.1315933 ,  0.13426131,
         0.13662781,  0.13874532,  0.14065379,  0.14238422,  0.14396119,
         0.0826078 ,  0.08704312,  0.09147844,  0.09591376,  0.10034907,
         0.1047844 ,  0.10921971,  0.11365503,  0.11809035,  0.12252566,
         0.126961  ,  0.1313963 ,  0.13583162,  0.14026694,  0.14470226,
        -0.01971253, -0.02020534, -0.02069815, -0.02119097, -0.02168378,
        -0.02217659, -0.0226694 , -0.02316222, -0.02365503, -0.02414784,
    ], dtype=np.float32)

    # add topo residual
    pbase, ts_topo_res = sim_pbase_and_topo_residual(
        num_date,
        delta_z_sim,
        ref_ind,
        max_pbase=max_pbase,
        inc_angle=inc_angle,
        range_dist=range_dist,
    )

    # observed time-series: displacement + topo res
    ts_obs = ts_sim + ts_topo_res

    # estimate DEM error
    G_defo = time_func.get_design_matrix4time_func(date_list, model)
    G_geom = np.reshape(pbase / (range_dist * np.sin(np.deg2rad(inc_angle))), (-1,1))
    G = np.hstack((G_geom, G_defo))

    delta_z_est, ts_cor = estimate_dem_error(
        ts_obs, G, tbase,
        phase_velocity=phase_velocity,
        cond=cond,
    )[:2]

    # plot
    if plot:
        plot_result(date_list, ts_sim, ts_obs, ts_cor, model)

    # validate
    print(f'Specified DEM error: {delta_z_sim:.2f} m')
    print(f'Estimated DEM error: {delta_z_est[0]:.2f} m')
    assert math.isclose(delta_z_sim, delta_z_est, rel_tol=rel_tol)
    print('Pass.')


################################################################################
def main(iargs=None):

    print('-'*50)
    print(f'Testing {__file__}')
    inps = cmd_line_parse(iargs)

    # prepare common data: date_list, tbase
    dt_list = [start_date + revisit_time * x for x in range(num_date)]
    date_list = [x.strftime('%Y%m%d') for x in dt_list]

    tbase = np.array(ptime.date_list2tbase(date_list)[0]) / 365.25   # years
    tbase -= tbase[ref_ind]

    # scenario 1 - simple linear deformation
    test_dem_error_with_linear_defo(
        date_list,
        tbase,
        rel_tol=0.05,
        plot=inps.plot,
    )

    # scenario 2 - complex nonlinear deformation (with more relaxed tolerance: 10%)
    test_dem_error_with_complex_defo(
        date_list,
        tbase,
        rel_tol=0.10,
        plot=inps.plot,
    )


################################################################################
if __name__ == '__main__':
    main(sys.argv[1:])



# L1-norm regularized least-squares solver, or LASSO (least absolute shrinkage and selection operator)
#
# Modified from CVXOPT-1.2 on GitHub (examples/doc/chap8/l1regls.py)
# by Zhang Yunjun, 10 Jan 2019 (add alpha argument)
#
# Another implementation is in cvxpy on GitHub (examples/lasso.py),
# which also support integer solution. But this implementation is very slow.
#
# Reference:
# Andersen, M., J. Dahl, Z. Liu, and L. Vandenberghe (2011), Interior-point methods for large-scale
# cone programming, in Optimization for machine learning, edited by S. Sra, S. Nowozin and S. J. Wright,
# MIT Press.


import math

from cvxopt import (
    blas,
    div,
    lapack,
    matrix,
    mul,
    normal,
    setseed,
    solvers,
    spdiag,
    sqrt,
)


def l1regls(A, y, alpha=1.0, show_progress=1):
    """

    Returns the solution of l1-norm regularized least-squares problem

        minimize || A*x - y ||_2^2  + alpha * || x ||_1.

    Parameters: A : 2D cvxopt.matrix for the design matrix in (m, n)
                y : 2D cvxopt.matrix for the observation in (m, 1)
                alpha : float for the degree of shrinkage
                show_progress : bool, show solving progress
    Returns:    x : 2D cvxopt.matrix in (m, 1)
    Example:    A = matrix(np.array(-C, dtype=float))
                b = matrix(np.array(closure_int, dtype=float).reshape(C.shape[0], -1))
                x = np.round(l1reg_lstsq(A, b, alpha=1e-2))
    """
    solvers.options['show_progress'] = show_progress

    m, n = A.size
    q = matrix(alpha, (2*n,1))
    q[:n] = -2.0 * A.T * y

    def P(u, v, alpha = 1.0, beta = 0.0 ):
        """
            v := alpha * 2.0 * [ A'*A, 0; 0, 0 ] * u + beta * v
        """
        v *= beta
        v[:n] += alpha * 2.0 * A.T * (A * u[:n])


    def G(u, v, alpha=1.0, beta=0.0, trans='N'):
        """
            v := alpha*[I, -I; -I, -I] * u + beta * v  (trans = 'N' or 'T')
        """

        v *= beta
        v[:n] += alpha*(u[:n] - u[n:])
        v[n:] += alpha*(-u[:n] - u[n:])

    h = matrix(0.0, (2*n,1))


    # Customized solver for the KKT system
    #
    #     [  2.0*A'*A  0    I      -I     ] [x[:n] ]     [bx[:n] ]
    #     [  0         0   -I      -I     ] [x[n:] ]  =  [bx[n:] ].
    #     [  I        -I   -D1^-1   0     ] [zl[:n]]     [bzl[:n]]
    #     [ -I        -I    0      -D2^-1 ] [zl[n:]]     [bzl[n:]]
    #
    # where D1 = W['di'][:n]**2, D2 = W['di'][n:]**2.
    #
    # We first eliminate zl and x[n:]:
    #
    #     ( 2*A'*A + 4*D1*D2*(D1+D2)^-1 ) * x[:n] =
    #         bx[:n] - (D2-D1)*(D1+D2)^-1 * bx[n:] +
    #         D1 * ( I + (D2-D1)*(D1+D2)^-1 ) * bzl[:n] -
    #         D2 * ( I - (D2-D1)*(D1+D2)^-1 ) * bzl[n:]
    #
    #     x[n:] = (D1+D2)^-1 * ( bx[n:] - D1*bzl[:n]  - D2*bzl[n:] )
    #         - (D2-D1)*(D1+D2)^-1 * x[:n]
    #
    #     zl[:n] = D1 * ( x[:n] - x[n:] - bzl[:n] )
    #     zl[n:] = D2 * (-x[:n] - x[n:] - bzl[n:] ).
    #
    # The first equation has the form
    #
    #     (A'*A + D)*x[:n]  =  rhs
    #
    # and is equivalent to
    #
    #     [ D    A' ] [ x:n] ]  = [ rhs ]
    #     [ A   -I  ] [ v    ]    [ 0   ].
    #
    # It can be solved as
    #
    #     ( A*D^-1*A' + I ) * v = A * D^-1 * rhs
    #     x[:n] = D^-1 * ( rhs - A'*v ).

    S = matrix(0.0, (m,m))
    Asc = matrix(0.0, (m,n))
    v = matrix(0.0, (m,1))

    def Fkkt(W):

        # Factor
        #
        #     S = A*D^-1*A' + I
        #
        # where D = 2*D1*D2*(D1+D2)^-1, D1 = d[:n]**-2, D2 = d[n:]**-2.

        d1, d2 = W['di'][:n]**2, W['di'][n:]**2

        # ds is square root of diagonal of D
        ds = math.sqrt(2.0) * div( mul( W['di'][:n], W['di'][n:]),
            sqrt(d1+d2) )
        d3 =  div(d2 - d1, d1 + d2)

        # Asc = A*diag(d)^-1/2
        Asc = A * spdiag(ds**-1)

        # S = I + A * D^-1 * A'
        blas.syrk(Asc, S)
        S[::m+1] += 1.0
        lapack.potrf(S)

        def g(x, y, z):

            x[:n] = 0.5 * ( x[:n] - mul(d3, x[n:]) +
                mul(d1, z[:n] + mul(d3, z[:n])) - mul(d2, z[n:] -
                mul(d3, z[n:])) )
            x[:n] = div( x[:n], ds)

            # Solve
            #
            #     S * v = 0.5 * A * D^-1 * ( bx[:n] -
            #         (D2-D1)*(D1+D2)^-1 * bx[n:] +
            #         D1 * ( I + (D2-D1)*(D1+D2)^-1 ) * bzl[:n] -
            #         D2 * ( I - (D2-D1)*(D1+D2)^-1 ) * bzl[n:] )

            blas.gemv(Asc, x, v)
            lapack.potrs(S, v)

            # x[:n] = D^-1 * ( rhs - A'*v ).
            blas.gemv(Asc, v, x, alpha=-1.0, beta=1.0, trans='T')
            x[:n] = div(x[:n], ds)

            # x[n:] = (D1+D2)^-1 * ( bx[n:] - D1*bzl[:n]  - D2*bzl[n:] )
            #         - (D2-D1)*(D1+D2)^-1 * x[:n]
            x[n:] = div( x[n:] - mul(d1, z[:n]) - mul(d2, z[n:]), d1+d2 )\
                - mul( d3, x[:n] )

            # zl[:n] = D1^1/2 * (  x[:n] - x[n:] - bzl[:n] )
            # zl[n:] = D2^1/2 * ( -x[:n] - x[n:] - bzl[n:] ).
            z[:n] = mul( W['di'][:n],  x[:n] - x[n:] - z[:n] )
            z[n:] = mul( W['di'][n:], -x[:n] - x[n:] - z[n:] )

        return g

    return solvers.coneqp(P, q, G, h, kktsolver = Fkkt)['x'][:n]


def test():
    m, n = 100, 1000
    setseed()
    A = normal(m,n)
    b = normal(m)
    x = l1regls(A, b)
    return x
